import pandas as pd
import numpy as np

from sklearn.utils.metaestimators import _safe_split
import sklearn.preprocessing as preprocessing
from sklearn.base import BaseEstimator, TransformerMixin

from pyadlml.dataset._representations.raw import create_raw, resample_raw
from pyadlml.dataset._representations.changepoint import create_changepoint, resample_changepoint
from pyadlml.dataset._representations.lastfired import create_lastfired, resample_last_fired
from pyadlml.dataset._core.acts_and_devs import label_data
from pyadlml.constants import ACTIVITY, TIME, DEVICE, VALUE, END_TIME, START_TIME, \
                              ENC_RAW, ENC_LF, ENC_CP, REPS
from pyadlml.dataset._core.devices import create_device_info_dict
from pyadlml.pipeline import XOrYTransformer, XAndYTransformer, YTransformer
from pyadlml.constants import OTHER


class Df2Numpy(TransformerMixin, XOrYTransformer):
    """ Transforms dataframes to numpy arrays by respecting
    the column order
    """
    _x_columns = []
    _y_columns = []

    def __init__(self):
        pass

    def fit(self, X, y=None):
        self._x_columns = X.columns
        if y is not None:
            self._y_columns = y.columns


    def fit_transform(self, X, y):
        self.fit(X,y)
        return self.transform(X,y)

    def transform(self, X, y):
        # reorder columns
        X = X[self._x_columns]
        y = y[self._y_columns]

        #
        X_np = X.values
        y_np = y.values
        return X_np, y_np


class DfCaster(TransformerMixin, XOrYTransformer):
    """ Transforms dataframes to torch tensors by respecting
    the column order
    """
    _x_columns = []
    _y_columns = []
    Df2Numpy = 'df->np'
    Df2Torch = 'df->torch'

    def __init__(self, x_conv, y_conv):
        self.y_conv = y_conv
        self.x_conv = x_conv

    def fit(self, X, y=None):
        assert self.y_conv in [self.Df2Numpy, self.Df2Torch]
        assert self.x_conv in [self.Df2Numpy, self.Df2Torch]

        self._x_columns = X.columns
        if y is not None:
            self._y_columns = y.columns


    def fit_transform(self, X, y=None):
        self.fit(X, y=y)
        return self.transform(X, y=y)

    def transform(self, X, y=None):
        """
        Parameters
        ----------
        X : pd
        """

        # reorder columns
        X = X[self._x_columns]

        # cast to respective format
        if self.x_conv == self.Df2Torch:
            import torch
            X_prime = torch.tensor(X.values, dtype=torch.float32)
        elif self.x_conv == self.Df2Numpy:
            X_prime = X.values.astype('float32')

        if y is not None:
            y = y[self._y_columns]
            if self.y_conv == self.Df2Torch:
                import torch
                y_prime = torch.tensor(y.values, dtype=torch.int64)
            elif self.y_conv == self.Df2Numpy:
                # ensure that y has dim (N,) and is of correct type
                y_prime = y.values.squeeze().astype('int64')
            return X_prime, y_prime
        else:
            return X_prime


class Df2Torch(TransformerMixin, XOrYTransformer):
    """ Transforms dataframes to torch tensors by respecting
    the column order
    """
    _x_columns = []
    _y_columns = []

    def __init__(self, only_y=False, only_X=False):
        self.only_y = only_y
        self.only_X = only_X

    def fit(self, X, y=None):
        self._x_columns = X.columns
        if y is not None:
            self._y_columns = y.columns


    def fit_transform(self, X, y):
        self.fit(X,y)
        return self.transform(X,y)

    def transform(self, X, y):
        """
        Parameters
        ----------
        X : pd
        """
        import torch

        # reorder columns
        X = X[self._x_columns]
        y = y[self._y_columns]

        # cast to tensors
        y_tensor = torch.tensor(y.values, dtype=torch.int64)
        X_tensor = torch.tensor(X.values, dtype=torch.float32)

        return X_tensor, y_tensor


class IdentityTransformer(TransformerMixin, XOrYTransformer):
    def __init__(self):
        pass

    def fit(self, X, y=None):
        return self

    @XOrYTransformer.x_or_y_transform
    def transform(self, X=None, y=None):
        return X, y


class RandomUnderSampler(TransformerMixin, XAndYTransformer):

    def __init__(self, sampling_strategy):
        self.sampling_strategy = sampling_strategy

    def fit(self, X, y, *args):
        from imblearn.under_sampling import RandomUnderSampler as ImbRUS
        self.rus_ = ImbRUS(sampling_strategy=self.sampling_strategy)
        return self

    def fit_transform(self, X, y, **fit_params):
        self.fit(X, y)
        return self.transform(X, y)

    def transform(self, X, y):
        return self.rus_.fit_resample(X,y)


class DropNanRows(XOrYTransformer):
    """
    Drops all rows where the label or the data has a nan-value
    """
    def fit(self, X,y=None):
        return self

    def fit_transform(self, X, y=None):
        return self.transform(X, y=y)

    def transform(self, X, y=None):
        """
        If X is given transform y
        """
        if y is not None:
            if isinstance(y, pd.DataFrame):
                nan_mask = y[ACTIVITY].isna().values
                return X[~nan_mask], y[~nan_mask]

            elif isinstance(y, np.ndarray):
                raise NotImplementedError
        else:
            nan_mask = X.isna().values
            return X[~nan_mask]


class DropDuplicates(TransformerMixin, XOrYTransformer):
    def __init__(self, ignore_columns=[], merge_on='time'):
        BaseEstimator.__init__(self)
        XOrYTransformer.__init__(self)

        self.ignore_columns = [ignore_columns] if not isinstance(ignore_columns, list) else ignore_columns
        self.merge_on = merge_on

    def fit(self, X, y=None):
        return self

    def fit_transform(self, X, y=None, **fit_params):
        return self.transform(X, y)

    @XOrYTransformer.x_or_y_transform
    def transform(self, X=None, y=None):
        """  Depending if X or y is given drop either all Nans in y or all nans in X or
        all Nans in the union of X and y.

        """
        if y is not None and X is not None:
            assert isinstance(y, pd.DataFrame)
            tmp = X.copy().reset_index(drop=True)
            y = y.copy().reset_index(drop=True)

            y_cols = y.columns.to_list()
            if len(y.columns) > 1:
                y_cols.remove(self.merge_on)
                tmp = tmp.merge(y, on=self.merge_on)
            else:
                tmp[y_cols] = y
            comp_cols = list(set(tmp.columns) - set(self.ignore_columns))
            dup_mask = tmp[comp_cols].duplicated()

            y = y[~dup_mask].reset_index(drop=True)
            X = X[~dup_mask].reset_index(drop=True)

        if y is not None and X is None:
            y = y.drop_duplicates()
        if X is not None and y is None:
            X = X.drop_duplicates()
        return X, y


class DropTimeIndex(TransformerMixin, XOrYTransformer):
    def __init__(self):
        XOrYTransformer.__init__(self)

    def fit(self, X=None, y=None):
        return self

    def fit_transform(self, X=None, y=None, **fit_params):
        """ Drops columns that are not time dependent
        """
        return self.transform(X, y)

    @XOrYTransformer.x_or_y_transform
    def transform(self, X=None, y=None):
        if X is not None:
            X = X.loc[:, X.columns != TIME]
        if y is not None:
            y = y.loc[:, y.columns != TIME]
        return X, y


class CVSubset(TransformerMixin, XOrYTransformer):
    """ Selects the subset from the whole dataset based on a given data_range.

    Attributes
    ----------
    data_range  : list of either indices or interval
        Either
    y : boolean, default=False
        Only

    """
    def __init__(self, data_range=None, y=False):
        self.data_range = data_range
        XOrYTransformer.__init__(self)

    def set_range(self, data_range):
        self.data_range = data_range

    def _extract_first_sample(self, data_range):
        try:
            sample = data_range[0][0]
        except TypeError:
            sample = data_range[0]
        except IndexError:
            sample = data_range[0]
        return sample


    def fit(self, X, y=None):
        error_msg = "Data range has to be set to either time intervals or list of indicies. Not 'None'"
        assert self.data_range is not None, error_msg

        # determine whether the splits are made based on index ranges or timestamp tuples
        sample = self._extract_first_sample(self.data_range)
        self.timestamps_ = not (isinstance(sample, int) or isinstance(sample, np.int64))

        return self

    def fit_transform(self, X, y=None):
        self.fit(X, y)
        return self.transform(X, y)

    @XOrYTransformer.x_or_y_transform
    def transform(self, X, y=None):
        """ Selects a subset of X and y based on the data_range and the split type
        """
        if self.timestamps_:
            x_mask = _create_mask(X, self.data_range)
            X = X[x_mask]
            if y is not None:
                assert len(X) == len(y)
                y_mask = _create_mask(y, self.data_range)
                y = y[y_mask]
            return X, y
        else:
            X, y = _safe_split(None, X, y, self.data_range)
            return X, y


def _create_mask(X, date_range):
    x_mask = (X[TIME] == 'false_init')
    if isinstance(date_range, tuple):
        date_range = [date_range]
    for time_pair in date_range:
        x_mask = x_mask | ((time_pair[0] < X[TIME]) & (X[TIME] < time_pair[1]))
    return x_mask


class StateVectorEncoder(BaseEstimator, TransformerMixin):
    """
    Create a State-vector sequence from a device event stream.
    Read more in the :ref:`User Guide <preprocessing_discretization>`.

    Parameters
    ----------
    encode : {'raw', 'changepoint', 'last_fired'}, default='raw'
        Determines the state-vectors encoding.

        raw
            Encode the event stream as vector, where each field
            represents a device. Each entry represents the current
            state that the device is in.

        changepoint
            Encode the event stream as one-hot binary vector, where
            each field represents a device. A one indicates the
            device that produced the event.

        last_fired
            Encode the event stream as one-hot binary vector,
            where each field represents a device. The device that produced
            the last event is indicated with one.

    dt : str, optional, default=None
        The timeslices resolution for discretizing the event stream. If
        set to None the event stream is not discretized.

    Attributes
    ----------
    encode : string
        The encoding for the vectors
    dt : str or None
        If not None, determines the binsize the event stream is discretized into
    data_info_ : dict or None
        Dictionary containing further device information such as most likely value
        or device datatype. This is used to speed up internal computation.
    classes_ : list
        A list of all devices. The lists order corresponds to the device order of the
        transformed state-vectors.

    Examples
    --------
    >>> from pyadlml.dataset import fetch_amsterdam
    >>> data = fetch_amsterdam()
    >>> from pyadlml.preprocessing import StateVectorEncoder
    >>> sve = StateVectorEncoder(encode='raw')
    >>> X = sve.fit_transform(data.df_devices)
    >>> sve2 = StateVectorEncoder(encode='changepoint+raw', dt='6s')
    >>> X_upsampled = sve2.fit_transform(data.df_devices)

    """

    def __init__(self, encode=ENC_RAW, dt=None):
        self.encode = encode
        self.dt = dt
        self.data_info_ = None
        self.classes_ = []

    @classmethod
    def _is_valid_encoding(cls, encoding):
        from itertools import chain, permutations
        def helper(iterable):
            s = list(iterable)
            return chain.from_iterable(permutations(s, r) for r in range(len(s)+1))
        # generate all possible valid combinations
        valid_combos = ['+'.join(combo) for combo in helper(REPS)]
        return encoding in valid_combos

    def fit(self, df_devs, y=None):
        """
        Fit the estimator.

        Parameters
        ----------
        df_devs : pd.DataFrame, optional
            recorded devices from a dataset. For more information refer to
            :ref:`user guide<device_dataframe>`.
        y : None
            Ignored. This parameter exists only for compatibility with
            :class:`~sklearn.pipeline.Pipeline`.

        Returns
        -------
        self
        """
        assert StateVectorEncoder._is_valid_encoding(self.encode)

        # create hashmap off all input features with mean for numerical
        # and most common value for binary or categorical features
        self.data_info_ = create_device_info_dict(df_devs)
        self.classes_ = self.data_info_.keys()
        return self

    def transform(self, df_devs=None,y=None, initial_states={}):
        """
        Discretize the data.

        Parameters
        ----------
        df_devs : array-like of shape (n_samples, n_features)
            The device dataframe to be transformed.
        y : None
            todo copy standard scipy sentence
        initial_states : dict
            A dictionary with device state mapping containing information about
            the last state the device held before the first event. If no dictionary 
            is given the states are infered by most likely values for 
            categorical and boolean devices.

        Returns
        -------
        Xt : {ndarray, sparse matrix}, dtype={np.float32, np.float64}
            Data in the binned space. Will be a sparse matrix if
            `self.encode='onehot'` and ndarray otherwise.
        """
        PRAEFIX_LF = 'lf_'
        PRAEFIX_CP = 'cp_'


        df_lst = []
        iters = self.encode.split('+')
        for enc in iters:
            if enc == ENC_RAW:
                data = create_raw(df_devs, self.data_info_, dev_pre_values=initial_states)
                if self.dt is not None:
                    data = resample_raw(data, df_dev=df_devs, dt=self.dt,
                                        most_likely_values=self.data_info_
                                        )

                # convert boolean data into integers (1,0)
                dev_bool = [dev for dev in self.data_info_.keys() if self.data_info_[dev]['dtype'] == 'boolean']
                data[dev_bool] = data[dev_bool].astype(int)

            elif enc == ENC_CP:
                data = create_changepoint(df_devs)
                if self.dt is not None:
                    data = resample_changepoint(data, self.dt)

                # set values that are missing in transform but were present when fitting to 0
                dev_diff = set(self.classes_) - set(data.columns)
                if len(dev_diff) > 0:
                    for dev in dev_diff:
                        data[dev] = 0

                # add prefix to make column names unique
                if len(iters) > 1:
                    data.columns = [TIME] + list(map(PRAEFIX_CP.__add__, data.columns[1:]))


            elif enc == ENC_LF:
                data = create_lastfired(df_devs)
                if self.dt is not None:
                    data = resample_last_fired(data, self.dt)

                # set values that are missing in transform but were present when fitting to 0
                dev_diff = set(self.classes_) - set(data.columns)
                if len(dev_diff) > 0:
                    for dev in dev_diff:
                        data[dev] = 0

                # add prefix to make column names unique
                if len(iters) > 1:
                    data.columns = [TIME] + list(map(PRAEFIX_LF.__add__, data.columns[1:]))

            else:
                raise ValueError
            data = data.set_index(TIME)
            df_lst.append(data)

        return pd.concat(df_lst, axis=1).reset_index()


    def fit_transform(self, df_devs, y=None):
        """

        Parameters
        ----------
        df_devs : pd.DataFrame, optional
            recorded devices from a dataset. For more information refer to
            :ref:`user guide<device_dataframe>`.
        y : None
            Ignored. This parameter exists only for compatibility with
            :class:`~sklearn.pipeline.Pipeline`.
        """
        self.fit(df_devs)
        return self.transform(df_devs)


class LabelMatcher(BaseEstimator, TransformerMixin, YTransformer):
    """
    Labels a dataframe with corresponding activities

    Attributes
    ----------
    idle : bool, default=False
        If true items that are not

    """
    def __init__(self, other : bool = False):
        """
        Parameters
        ----------
        other : bool, default=False
            The other activity TODO

        Returns
        -------
        self
        """
        self.other = other

    def fit(self, y, X):
        """
        Fit label encoder.

        Parameters
        ----------
        y : pd.DataFrame
            recorded activities from a dataset. Fore more information refer to the
            :ref:`user guide<activity_dataframe>`.

        Returns
        -------
        self : returns an instance of self
        """
        self.lbl_enc_ = preprocessing.LabelEncoder()
        self.df_acts_ = y

        # add actvities to sklearns LabelEncoder
        activities = list(y[ACTIVITY].values)
        if self.other:
            activities.append(OTHER)
        #self.lbl_enc_.fit(activities)
        #self.classes_ = self.lbl_enc_.classes_
        self.classes_ = activities
        return self


    def fit_transform(self, y, X):
        """
        Fit label encoder and return encoded labels.

        Parameters
        ----------
        y : pd.DataFrame
            An activity dataframe recorded activities from a dataset. Fore more information refer to the
            :ref:`user guide<activity_dataframe>`.
        X : pd.DataFrame
            A dataframe containing a column with name 'time' of timestamps.

        Returns
        ------
        df : pd.DataFrame
            The X dataframe with an additional column containing the corresponding activity labels
        """
        self.fit(y, X)
        return self.transform(y, X)

    def inverse_transform(self, y, retain_index=False):
        """
        Transform labels back to original encoding.

        Parameters
        ----------
        x : array like or pd.DataFrame or pd.Series
            array of numbers that are transformed to labels
        retain_index : bool, default=False
            TODO

        Returns
        -------
            TODO
        """
        if isinstance(y, int):
            # Case when a string is passed
            return self.id2class(y)
        elif self._is_iterable(y):
            # Iter through thing and replace values with encodings
            return self._wrap_iterable(list(map(self.id2class, y)), y)
        else:
            raise ValueError('Passed weird argument!!!')
        raise NotImplementedError
        #if isinstance(x, np.ndarray):
        #    res = self._lbl_enc.inverse_transform(x)
        #    if retain_index:
        #        return pd.Series(data=res, index=self.df_devs.index[:len(res)])
        #    else:
        #        return res
        #
        #elif isinstance(x, pd.DataFrame) or isinstance(x, pd.Series):
        #    tmp_index = x.index
        #    res = self._lbl_enc.inverse_transform(x.values)
        #    return pd.Series(data=res, index=tmp_index)
        #else:
        #    raise ValueError

    def id2class(self, y: int) -> str:
        """ Get class label corresponding to the numeric label.
        """
        return self.classes_[y]


    def class2id(self, y: str) -> int:
        """ Get numeric label from class label.
        """
        return np.argmax(self.classes_ == y)

        
    def _is_iterable(self, z):
        try:
            iter(z)
            return True
        except:
            return False


    def _wrap_iterable(self, lst, iter):
        """ list 
        """
        if isinstance(iter, np.ndarray):
            return np.array(lst)
        elif isinstance(iter, list):
            return lst
        elif isinstance(iter, list):
            return lst
        elif isinstance(iter, pd.Series):
            tmp = pd.Series(lst)
            tmp.index = iter.index
            return tmp
        else:
            raise NotImplementedError


    def transform(self, y: pd.DataFrame, X: pd.DataFrame = None) -> pd.DataFrame:
        """

        Parameters
        ----------
        y : pd.DataFrame
            An activity dataframe
        X : pd.DataFrame
            A device dataframe

        Returns
        -------
        pd.DataFrame
            A dataframe with column TIME, ACTIVITY matchin the rows of X
            TODO asdf
        """
        # normal case where X an y are provided
        if X is not None:
            return label_data(X, y, self.other)[[TIME, ACTIVITY]].copy()
            # remove Nans before encoding the labels and then concatenate again
            #if not self.other:
                #nan_mask = df[ACTIVITY].isna()
                #df_san_nan = df[~nan_mask]
                #encoded_labels = self.lbl_enc_.transform(df_san_nan[ACTIVITY].values)
                #new_san_nan = pd.DataFrame(data={TIME: df_san_nan[TIME].values, ACTIVITY: encoded_labels})
                #new_nans = pd.DataFrame(data={TIME: df[nan_mask][TIME].values, ACTIVITY: np.nan})
                #return pd.concat([new_nans, new_san_nan]).sort_values(by=TIME)
            #else:
                #encoded_labels = self.lbl_enc_.transform(df[ACTIVITY].values)
                #return pd.DataFrame(data={TIME: df[TIME].values, ACTIVITY: encoded_labels})
        else:

            if isinstance(y, str):
                # Case when a string is passed
                return self.class2id(y)
            elif self._is_iterable(y):
                # Iter through thing and replace values with encodings
                return self._wrap_iterable(list(map(self.class2id, y)), y)
            else:
                raise ValueError('Passed weird argument!!!')


class SequenceSlicer(TransformerMixin, XOrYTransformer):
    """ Slices a sequence into batches.

    Parameters
    ----------

    .. image:: ../_static/images/many_to_many.svg
       :height: 200px
       :width: 500 px
       :scale: 90%
       :alt: alternate text
       :align: center


    .. image:: ../_static/images/reps/image.svg
       :height: 200px
       :width: 500 px
       :scale: 80%
       :alt: alternate text
       :align: center


    Many-To-Many
    ^^^^^^^^^^^^

    To get *many-to-many* batches use the window size :math:`w` to split the data

    .. math::
        f(X_{N,K},y_{N}) \rightarrow (X_{W, N,K}, y_{W, N})

    In addition you can specify a stride.


    .. code:: python

        from pyadlml.preprocessing import SequenceDicer

        raw = StateVectorEncoder(encode='raw', t_res='10s')\
              .fit_transform(data.df_devices)
        labels = LabelEncoder().fit_transform(raw, data.df_activities)

        X = raw.values
        y = labels.values

        X, y = SequenceDicer(rep='many-to-many', window_length='5m')\
               .fit_transform(X, y)

    Many-To-One
    ^^^^^^^^^^^

    .. math::
        f(X_{N,K},y_{N}) \rightarrow (X_{W, N, K}, y_{N})


    .. code:: python

        from pyadlml.preprocessing import SequenceDicer

        raw = StateVectorEncoder(encode='raw', t_res='10s')\
              .fit_transform(data.df_devices)
        labels = LabelEncoder().fit_transform(raw, data.df_activities)

        X = raw.values
        y = labels.values

        X, y = SequenceDicer(rep='many-to-one', window_length='20s')\
               .fit_transform(X, y)

    """
    REP_M2M = 'many-to-many'
    REP_M2O = 'many-to-one'

    def __init__(self, rep='many-to-many', window_size=10, stride=1):
        TransformerMixin.__init__(self)
        XOrYTransformer.__init__(self)

        self.rep = rep
        self.window_size = window_size
        self.stride = stride

    def fit(self, X,y=None, **fit_params):
        # TODO not having this method throws an error ! dirty hack
        return self

    def fit_transform(self, X, y=None):
        """
        Parameters
        ----------
        X : nd.array
            Some kind of numpy array
        y : nd.array

        Returns
        -------
        x : nd.array
            todo
        y : nd.array
            todo
        """
        return self.transform(X, y)

    @XOrYTransformer.x_or_y_transform
    def transform(self, X=None, y=None):
        """

        """
        assert self.rep in [self.REP_M2M, self.REP_M2O]

        if X is not None:
            assert self.window_size < len(X)
            assert self.stride < len(X)

            X = self._transform_X(X)

        if y is not None:
            assert self.window_size < len(y)
            assert self.stride < len(y)

            y = self._transform_Y(y)

        return X, y

    def _calc_new_N(self, n_old):
        return int(np.floor((n_old-self.window_size)/self.stride)+1)

    def _transform_X(self, X):
        """

        """
        n_prime = self._calc_new_N(X.shape[0])
        new_shape =[n_prime, self.window_size, X.shape[1]]

        res = np.zeros(shape=new_shape, dtype='float32')
        if not isinstance(X, np.ndarray):
            import torch
            if isinstance(X, torch.Tensor):
                try:
                    res = torch.zeros(new_shape, dtype=torch.float32)
                except TypeError:
                    print('asdf')
            else:
                raise ValueError('X has to be either ndarray or of type torch tensor')

        for r, i in enumerate(range(0, n_prime*self.stride, self.stride)):
            res[r, :, :] = X[i: i+self.window_size, :]
        return res

    def _transform_Y(self, y):
        """


        """
        dtype = 'int64'
        if self.rep == self.REP_M2M:
            n_prime = self._calc_new_N(y.shape[0])
            new_shape = [n_prime, self.window_size]

            if isinstance(y, np.ndarray):
                res = np.zeros(shape=[n_prime, self.window_size], dtype=object)
                res = res.astype(dtype)
            else:
                import torch
                if isinstance(y, torch.Tensor):
                    res = torch.zeros(new_shape, dtype=torch.int64)
                else:
                    raise ValueError('X has to be either ndarray or of type torch tensor')

            for r, i in enumerate(range(0, n_prime*self.stride, self.stride)):
                res[r, :] = y[i: i+self.window_size]
            return res.squeeze()

        elif self.rep == self.REP_M2O:
            res = y[list(range(self.window_size-1, y.shape[0], self.stride))]
            if isinstance(y, np.ndarray):
                res = res.squeeze().astype(dtype)
                return res
            else:
                import torch
                raise NotImplementedError
                #return res
        else:
            raise ValueError


import os
import pandas as pd
from pathlib import Path
from pyadlml.constants import START_TIME, END_TIME, DEVICE, VALUE, TIME, ACTIVITY
from pyadlml.dataset._core.activities import ActivityDict

DATA_NAME = 'devices.csv'
DEV_MAP_NAME = 'device_map.csv'
ACT_MAP_NAME = 'activity_map.csv'
ACT_NAME = 'activities_subject_%s.csv'
MAP_ID = 'id'

def _read_activities(path_to_file: Path, path_to_mapping: dict):
    """
    """
    activities = pd.read_csv(path_to_file)
    act_map = pd.read_csv(path_to_mapping, index_col=MAP_ID)\
                .to_dict()[ACTIVITY]
    activities[ACTIVITY] = activities[ACTIVITY].map(act_map)
    activities[START_TIME] = pd.to_datetime(activities[START_TIME])
    activities[END_TIME] = pd.to_datetime(activities[END_TIME])
    return activities


def _read_devices(path_to_dev_file, path_to_mapping):
    devices = pd.read_csv(path_to_dev_file)
    dev_map = pd.read_csv(path_to_mapping, index_col=MAP_ID)\
                .to_dict()[DEVICE]

    devices[DEVICE] = devices[DEVICE].map(dev_map)
    devices[VALUE] = devices[VALUE].astype(str).replace({
        "0": False, "0.0": False, 
        "1": True, "1.0": True,
        "True":True, "False": False,
        "true":True, "false": False,
    })
    devices[TIME] = pd.to_datetime(devices[TIME])
    devices = devices.reset_index(drop=True)
    return devices


def _read_activity_list(path_to_file):
    lst_activities = pd.read_csv(path_to_file)
    return list(lst_activities[ACTIVITY])


def _read_device_list(path_to_file):
    lst_devices = pd.read_csv(path_to_file)
    return list(lst_devices[DEVICE])


def load(folder_path: str, subjects:list=[], retain_corrections=False) -> dict:
    """
    Loads a dataset generated by activity-assistant from a specified folder.

    Parameters
    ----------
    folder_path : str
        The path to the folder where the dataset is located.

    subjects : list of str
        The subjecst names that are to be included. If the activity file
        is named *activities_subject_foo.csv*, then *foo* is the corresponding name.

    Examples
    --------
    The dataset ``my_dataset.zip`` is extracted in the folder ``/path/to/``.

    >>> from pyadlml.dataset import load_act_assist
    >>> data = load_act_assist('/path/to/my_dataset/', subjects=['chris'])

    Returns
    -------
    data : object
    """
    assert isinstance(folder_path, str) or isinstance(folder_path, Path)
    assert isinstance(subjects, list)

    df_dev = _read_devices(Path(folder_path).joinpath(DATA_NAME),
                           Path(folder_path).joinpath(DEV_MAP_NAME))

    # get mappings
    lst_dev = _read_device_list(Path(folder_path).joinpath(DEV_MAP_NAME))
    lst_act = _read_activity_list(Path(folder_path).joinpath(ACT_MAP_NAME))

    #data = Data(None, df_dev, activity_list=lst_act, device_list=lst_dev)



    # If no subjects where specified read the files from the folder following the naming schemata
    if not subjects:
        for fp in Path(folder_path).iterdir():
            if ACT_NAME[:18] in fp.name:
                subjects.append(fp.name[19:-4])

    act_dct = ActivityDict()
    for subject in subjects:
        df_act = _read_activities(Path(folder_path).joinpath(ACT_NAME%(subject)),
                                  Path(folder_path).joinpath(ACT_MAP_NAME))
        act_dct[subject] = df_act

    return dict(
        devices=df_dev,
        device_list=lst_dev,
        activities=act_dct,
        activity_list=lst_act,
    )



import dash
import dash.dcc as dcc
import numpy as np
import dash_daq as daq
import dash.html as html
import pandas as pd
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
from pyadlml.dataset.plot.plotly.activities import bar_count as act_bar_count, \
    boxplot_duration as act_boxplot_duration, density as act_density, \
    heatmap_transitions as act_heatmap_transitions
from pyadlml.dataset.plot.plotly.acts_and_devs import activity_vs_device_events_hist, event_correlogram
from pyadlml.dataset.plot.plotly.devices import bar_count as dev_bar_count, \
    device_iei as dev_iei, fraction as dev_fraction, event_density, boxplot_state, \
    plotly_event_correlogram as dev_cc
from pyadlml.dataset.plot.plotly.acts_and_devs import *
from dash.dependencies import *
from plotly.graph_objects import Figure

from pyadlml.constants import TIME, START_TIME, END_TIME, ACTIVITY, DEVICE
from pyadlml.dataset._core.activities import ActivityDict
from pyadlml.dataset.util import select_timespan, timestamp_to_num, num_to_timestamp

DEV_DENS_SLIDER = {i: e for i, e in enumerate(
    ['10s', '30s', '1min', '5min', '30min', '1h', '2h', '6h', '12h']
)}

LAG_SLIDER = {i: e for i, e in enumerate(
    ['10s', '30s', '1min', '2min', '5min', '30min', '1h', '2h']
)}

BIN_SIZE_SLIDER = {i: e for i, e in enumerate(
    ['1s', '5s', '10s', '20s', '30s', '1min', '2min', '5min', '10min']
)}


def _np_dt_strftime(ts, format):
    return pd.to_datetime(ts).strftime(format)


def _build_log_radio_buttons(id):
    return dcc.RadioItems(
        id=id,
        options=[{'label': i, 'value': i} for i in ['linear', 'log']],
        value='linear',
        labelStyle={'display': 'inline-block', 'marginTop': '5px'}
    )


def _build_row_radio_button(rd_id, options, value):
    """ Builds an inline radio button menu """
    return dbc.RadioItems(id=rd_id,
               inline=True,
               options=[{'label': i, 'value': i} for i in options],
               value=value,
               #labelStyle={'display': 'inline-block', 'marginTop': '5px'}
    )

def _build_sort_radio_buttons(id):
    return dcc.RadioItems(
        id=id,
        options=[{'label': i, 'value': i} for i in ['alphabetical', 'value', 'area']],
        value='value',
        labelStyle={'display': 'inline-block', 'marginTop': '5px'}
    )


def _build_choice_activity(df_acts, id, sel=None, multi=True):
    def create_option(name):
        return {'label': name, 'value': name}
    df_acts =  ActivityDict.wrap(df_acts)
    acts = df_acts.get_activity_union()
    if sel is None:
        act_sel = acts
    elif isinstance(sel, str) and multi:
        act_sel = np.array([sel])
    else:
        act_sel = sel

    options = [create_option(a) for a in acts]
    options = [create_option('other'), *options]
    return dcc.Dropdown(id=id,
                        multi=multi,
                        options=options,
                        value=act_sel,
                        clearable=False,
                        )


def _build_choice_devices(df_devs, id, sel=None, multi=True):
    def create_option(name):
        return {'label': name, 'value': name}
    devs = df_devs[DEVICE].unique()
    if sel is None:
        devs_sel = devs
    elif isinstance(sel, str) and multi:
        devs_sel = np.array([sel])
    else:
        devs_sel = sel
    options = [create_option(dev) for dev in devs]
    return dcc.Dropdown(
        id=id,
        options=options,
        multi=multi,
        value=devs_sel,
        clearable=False,
    )


def _build_range_slider(df_acts, df_devs, start_time, end_time, set_end_time):
    def create_mark(ts, format):
        return {'label': _np_dt_strftime(ts, format), 'style': {"transform": "rotate(45deg)"}}

    marks = {0: create_mark(start_time, '%Y.%m.%d'),
             1: create_mark(end_time, '%Y.%m.%d')}
    #marks[0]['style']['margin'] = '10px 0 0 0'      # top right bottom left
    #marks[1]['style']['margin'] = '10px 15px 0 0'    # top right bottom left


    # Determine the marker frequency
    diff = end_time - start_time
    if diff < pd.Timedelta('8W'):       # 2 Month -> daily frequency
        # Dataset Amsterdam, Mitlab_1/2W, Mitlab_2/2W, Aras/4W, UCIA/2W
        # UCIA/3W
        rng = pd.date_range(start_time, end_time, freq='D')[1:-1]
    elif diff < pd.Timedelta('16W'):    # 4 Months
        # Dataset Tuebingen2019
        rng = pd.date_range(start_time, end_time, freq='3D')[1:-1]
    elif diff < pd.Timedelta('32W'):    # 8 Months
        # Dataset Casas Aruba
        rng = pd.date_range(start_time, end_time, freq='W')[1:-1]
    else:
        rng = pd.date_range(start_time, end_time, freq='2W')[1:-1]

    for day in rng:
        marks[timestamp_to_num(day, start_time, end_time)] = create_mark(day, '%m.%d')
    set_end_time = timestamp_to_num(set_end_time, start_time, end_time)
    return dcc.RangeSlider(id='range-slider', min=0, max=1, step=0.001,
                           value=[0, set_end_time], marks=marks)


def _build_plot_header(h_id, title, op_id):
    """ Creates a header for a graph """
    return dbc.Row(
        dbc.Col(children=[
                    html.H6(title, id=h_id),
                    dbc.Button('options', id=op_id, color='link')
        ])
    )

def _build_options_btn(op_id):
    """ Creates options button to link to graph options"""
    return dbc.Row(justify='end',
                   children=dbc.Col(width=3,
                                    children=dbc.Button('options', style={'marginTop': '-18px'},
                                                        id=op_id, color='link', size='sm')))

def acts_vs_devs_layout(df_acts, act_id, df_devs, plot_height=False):
    """


    Note
    ----
    Since the contingency table calculation may take very long the layout is initalized 
    with placeholders and the real figures are computed later on demand
    """
    from copy import deepcopy
    df_acts = deepcopy(df_acts)

    activities = df_acts[ACTIVITY].unique()
    devices = df_devs[DEVICE].unique()

    cc_act_selected = np.random.choice(activities, 3).tolist()
    cc_dev_selected = np.random.choice(devices, 4).tolist()

    fig_cc = event_correlogram(
        df_devs, 
        df_acts, 
        to=cc_dev_selected,
        maxlag='1min', 
        binsize='2s'
    )

    fig_avd_hist = activity_vs_device_events_hist(
        df_devs,
        df_acts,
        device=cc_dev_selected[0],
        activity=cc_act_selected[0],
        normalize=True,
        height=250,
    )

    layout_acts_vs_devs = dbc.Container([
        dcc.Store(f'avd_activity-order-{act_id}'),
        dcc.Store(f'avd_device-order-{act_id}'),
        dcc.Store(f'avd_state-contingency-{act_id}'),
        dcc.Store(f'avd_event-contingency-{act_id}'),
        html.Div(id=f"avd-trigger-{act_id}", style=dict(display="none")),
        html.Div(id=f"avd-update-{act_id}", style=dict(display="none")),
        dbc.Row([
            dcc.Graph(id=f'avd_graph-event-contingency-{act_id}',
                      style=dict(height='100%',width='100%'),
                      figure=Figure(),
                      config=dict(displaylogo=False, displayModeBar=True,
                                  responsive=False,
                                  modeBarButtonsToRemove=_buttons_to_use('resetScale2d'),
                     ),
            ),
            _build_options_btn(f'clps-avd-event-button-{act_id}'),
            dbc.Collapse(id=f'clps-avd-event-{act_id}',
                        style={'paddingLeft': '2rem',
                                'paddingBottom': '2rem',
                    },
                children=_option_grid(
                     labels=['Scale:', 'Activity order: ', 'Device order'],
                     values=[
                        _build_row_radio_button(
                            rd_id=f'avd_event-scale-{act_id}',
                            options=['linear', 'log'],
                            value='log'),
                        _build_row_radio_button(
                            rd_id=f'avd_act-order-trigger-{act_id}',
                            options=['alphabetical', 'duration', 'count', 'area'],
                            value='alphabetical'),
                        _build_row_radio_button(
                            rd_id=f'avd_dev-order-trigger-{act_id}',
                            options=['alphabetical', 'count', 'area'],
                            value='alphabetical'),
                ])),
        ]),
        dbc.Row([
            dbc.Spinner(html.Div(id=f'loading-output-{act_id}')),
            dcc.Graph(id=f'avd_graph-state-contingency-{act_id}',
                      style=dict(height='100%',width='100%'),
                      figure=Figure(),
                      config=dict(displaylogo=False, displayModeBar=True)
            ),
            _build_options_btn(f'clps-avd-state-button-{act_id}'),
            dbc.Collapse(id=f'clps-avd-state-{act_id}',
             style={'paddingLeft': '2rem',
                    'paddingBottom': '2rem',
                    },
            children=_option_grid(
                 labels=['Scale:', 'Activity order: ', 'Device order'],
                 values=[
                    _build_row_radio_button(
                        rd_id=f'avd_state-scale-{act_id}',
                        options=['linear', 'log'],
                        value='log'),
            ])),
        ]),
        dbc.Row(
            children=[
                dbc.Col(width=12, children=[
                            dbc.Row(dcc.Graph(id=f'avd_graph-cc-{act_id}',
                                            figure=fig_cc,
                                            style=dict(height='100%',width='100%'),
                                            config=dict(displaylogo=False,
                                                    displayModeBar=True,
                                                    modeBarButtonsToRemove=_buttons_to_use('resetScale2d'),
                                                )
                            )),
                            _build_options_btn(f'clps-avd-cc-button-{act_id}'),
                            dbc.Collapse(id=f'clps-avd-cc-{act_id}',
                                        style={'paddingLeft': '2rem',
                                                'paddingRight': '2rem',
                                                'paddingBottom': '2rem',
                                                },
                                        children=_option_grid(
                                             labels=['Fix:', 'To:', 'Max lag:', 'Binsize:'],
                                             lwidth=3, rwidth=9,
                                             values=[
                                                _build_choice_activity(df_acts, id=f'avd_cc-sel-fix-{act_id}', sel=cc_act_selected),
                                                _build_choice_devices(df_devs, id=f'avd_cc-sel-to-{act_id}', sel=cc_dev_selected),
                                                dcc.Slider(id=f'avd_cc-lag-slider-{act_id}', min=0, max=len(LAG_SLIDER), step=None, value=3,
                                                        marks={i:{'label':LAG_SLIDER[i], 'style': {"transform": "rotate(45deg)"}} for i in range(0, len(LAG_SLIDER))}
                                                ),
                                                dcc.Slider(id=f'avd_cc-binsize-slider-{act_id}', min=0, max=len(BIN_SIZE_SLIDER), step=None, value=3,
                                                        marks={i:{'label':BIN_SIZE_SLIDER[i], 'style': {"transform": "rotate(45deg)"}} for i in range(0, len(BIN_SIZE_SLIDER))}
                                                ),
                                            ]
                                        )
                            )
                ]),
        ]),
        dbc.Row([
                dbc.Col(width=12, children=[
                            dbc.Row(dcc.Graph(id=f'avd_graph-hist-{act_id}',
                                          figure=fig_avd_hist,
                                          style=dict(height='100%',width='100%'),
                                          config=dict(displaylogo=False,
                                                displayModeBar=True,
                                                modeBarButtonsToRemove=_buttons_to_use(
                                                        'zoom2d', 'pan2d',  'lasso2d', 'resetScale2d'
                                                ),
                                          )
                                ),
                            ),
                            _build_options_btn(f'clps-avd-hist-button-{act_id}'),
                            dbc.Collapse(id=f'clps-avd-hist-{act_id}',
                                        style={'paddingLeft': '2rem',
                                            'paddingRight': '2rem',
                                            'paddingBottom': '2rem',
                                            },
                                    children=_option_grid(
                                            labels=['Normalize:', 'Device:', 'Activity: '],
                                            lwidth=4, rwidth=8,
                                            values=[
                                            _build_row_radio_button(
                                                    rd_id=f'avd_hist-normalize-{act_id}',
                                                    options=['True', 'False'],
                                                    value='True'),
                                            _build_choice_devices(df_devs, id=f'avd_hist-sel-dev-{act_id}', sel=cc_dev_selected[0], multi=False),
                                            _build_choice_activity(df_acts, id=f'avd_hist-sel-act-{act_id}', sel=cc_act_selected[0], multi=False),
                            ]))
                ]),
        ])
    ])
    return layout_acts_vs_devs

def _buttons_to_use(*use):
    """Returns the buttons to remove in order to get the buttons to use"""
    buttons = ['toImage', 'zoom2d', 'pan2d',  'select2d', 'lasso2d', 'zoomIn2d', 'zoomOut2d',
               'autoScale2d', 'resetScale2d',
               'toggleSpikelines', 'hoverCompareCartesian', 'hoverClosestCartesian'
    ]
    res = []
    for b in buttons:
        if b not in use:
            res.append(b)
    return res

def acts_n_devs_layout(df_devs, df_acts, start_time, end_time, set_end_time, plot_height=350):
    """
    TODO

    """
    if df_acts is None or df_devs is None:
        fig_and = Figure()
        time_slider = dcc.RangeSlider(id='range-slider', min=0, max=1, step=0.001, value=[0, 1])
        checklist_act = dcc.Dropdown(id='select-activities', multi=True, options=[], value=[], clearable=False)
        choice_device = dcc.Dropdown(id='select-devices', options=[], multi=True, value=[], clearable=False)
    else:
        fig_and = activities_and_devices(df_devs, df_acts, st=start_time, et=set_end_time, height=plot_height)
        time_slider = _build_range_slider(None, None, start_time, end_time, set_end_time)
        checklist_act = _build_choice_activity(df_acts, id='select-activities')
        choice_device = _build_choice_devices(df_devs, id='select-devices')

    return html.Div(children=[

                   dbc.Row(dbc.Col(
                            dcc.Graph(id='graph-acts_n_devs',
                                      style=dict(height='100%',width='100%'),
                                      figure=fig_and,
                                      config=dict(displayModeBar=True,
                                                  modeBarButtonsToRemove=_buttons_to_use(
                                                      'zoom2d', 'zoomIn2d', 'zoomOut2d',
                                                      'pan2d', 'resetScale2d'

                                                  ),
                                                  edits=dict(legendPosition=True),
                                                  showAxisDragHandles=True,
                                                  displaylogo=False))
                   )),
            dbc.Row(justify='end', children=dbc.Col(width=5,
                            children=[dbc.ButtonGroup([
                                dbc.Button('reset selection', id='and_reset_sel', disabled=True, color='link', size='sm'),
                                dbc.Button('options',
                                   id='clps-acts-n-devs-button',
                                   color='link',
                                   size='sm',
                                   n_clicks=0),
                                dbc.Button('cr', id='and_btn_copy_range', color='link', style={'fontsize': 10}),
                                dcc.Clipboard(id='and_clipboard', style={'fontsize': 10}),
                            ]),
                            ])),
            dcc.Store(id='and_act-order'),
            dcc.Store(id='and_dev-order'),
            dbc.Collapse(id='clps-acts-n-devs',
                children=[
                    html.H6('Time: '),
                    html.Div(children=[time_slider]),
                    html.H6('Activities: ', style={'marginTop': '20px'}),
                    dbc.Row(style={'padding': '10px'},
                            children=[
                            dbc.Row(children=[
                                dbc.Col(children=[
                                    dbc.Row(children=[
                                        dbc.Col('Order: ', width=2),
                                        dbc.Col(
                                            dbc.RadioItems(id='and_act-order-trigger',
                                                           inline=True,
                                                           options=[{'label': i, 'value': i} for i in ['alphabetical', 'count', 'duration']],
                                                           value='duration',
                                                           labelStyle={'display': 'inline-block', 'marginTop': '5px'}),
                                            width=10,
                                        )
                                        ],
                                    )],
                                        width=10,
                                ),
                                ]
                            ),
                            dbc.Row(children=[checklist_act]),
                    ]),
                    html.H6('Devices', style={'marginTop': '20px'}),
                    dbc.Row(style={'padding': '10px'},
                            children=[
                                dbc.Row(children=[
                                    dbc.Col(children=[
                                        dbc.Row(children=[
                                            dbc.Col('Order: ', width=2),
                                            dbc.Col(width=10, children=dbc.RadioItems(id='and_dev-order-trigger',
                                                               inline=True,
                                                               options=[{'label': i, 'value': i} for i in ['alphabetical', 'type', 'area']],
                                                               value='alphabetical',
                                                               labelStyle={'display': 'inline-block', 'marginTop': '5px'}),

                                            )
                                        ])
                                    ]),
                                    dbc.Row(children=[
                                            dbc.Col('Marker type: ', width=2),
                                            dbc.Col(width=10, children=dbc.RadioItems(id='and_dev-type',
                                                               inline=True,
                                                               options=[{'label': i, 'value': i} for i in ['event', 'state']],
                                                               value='event',
                                                               labelStyle={'display': 'inline-block', 'marginTop': '5px'}),

                                            )
                                        ])
                                ]),
                                dbc.Row(choice_device),
                    ]),
            ]),
    ])

def device_layout_graph_bottom(fig, type='density'):
    options_density = [_build_row_radio_button(
                rd_id='devs_iei-scale',
                options=['linear', 'log'],
                value='log'),
            daq.BooleanSwitch(
                id='devs_iei-per-device',
                on=False,
                persisted_props=[]
            )]
    options_iei = [
        dbc.Collapse(id='clps-dev-iei'),
        dbc.Button(id='clps-dev-iei-button'),
    ]

    if type == 'density':
        mode_bar_buttons = _buttons_to_use('zoom2d', 'pan2d', 'resetScale2d'),
        labels = []
        options = options_density
        to_hide = options_iei
    elif type == 'iei':
        mode_bar_buttons = _buttons_to_use('zoom2d', 'pan2d', 'resetScale2d'),
        labels = ['Scale:', 'Per device: ']
        options = options_iei
        to_hide = options_iei
    else:
        raise ValueError

    return [dbc.Row(dcc.Graph(figure=fig,
                              style=dict(height='100%',width='100%'),
                              config=dict(displaylogo=False,
                                    displayModeBar=True,
                                    modeBarButtonsToRemove=mode_bar_buttons
                              )
                        )),
                        _build_options_btn('clps-dev-iei-button'),
                        dbc.Collapse(id='clps-dev-iei',
                                     style={'paddingLeft': '2rem',
                                            'paddingRight': '2rem',
                                            'paddingBottom': '2rem',
                                            },
                                    children=_option_grid(
                                         labels=['Plot', *labels],
                                         lwidth=4, rwidth=8,
                                         values=[
                                            dcc.Dropdown(id='devs_bottom-drop',
                                                         value='density', options=[
                                                              {'label': 'Inter-event-interval', 'value': 'iei' },
                                                              {'label': 'density', 'value': 'density' }
                                                ]
                                            ),
                                             *options]
                                    )
                        ),
            html.Div(children=to_hide, style={'display': 'none'})
            ]


def devices_layout(df_devs, initialize=False, plot_height=350):

    cc_dev_selected = np.random.choice(df_devs[DEVICE].unique(), 4).tolist()

    if df_devs is not None and not df_devs.empty:
        order = 'count'
        fig_bar_count = dev_bar_count(df_devs, height=plot_height)
        fig_bp_state = boxplot_state(df_devs, height=plot_height, scale='log', order=order)
        fig_ev_density = event_density(df_dev=df_devs, show_colorbar=False,
                                    height=plot_height, order=order
                                    )
        fig_iei = dev_iei(df_devs, height=plot_height)
        fig_fraction = dev_fraction(df_devs, height=plot_height, order=order)

        # Choose random first device for presentation
        fig_cc = dev_cc(df_devs, height=plot_height, fix=cc_dev_selected, to=cc_dev_selected)
    else:
        fig_bar_count = Figure()
        fig_bp_state = Figure()
        fig_ev_density = Figure()
        fig_iei = Figure()
        fig_cc = Figure()
        fig_fraction = Figure()

    layout_devices = dbc.Container([
        dcc.Store('devs_density-data'),
        dcc.Store('devs_order'),
        html.Div(id="dev-trigger", style=dict(display="none")),
        html.Div(children='', id=f"dev-curr-sel", style=dict(display="none")),
        dcc.Store(id=f'dev-curr-sel-store'),
        dbc.Row([
            dbc.Col(width=6, children=[
                        dbc.Row(dcc.Graph(id='devs_graph-bar',
                                          figure=fig_bar_count,
                                          style=dict(height='100%',width='100%'),
                                          config=dict(displaylogo=False,
                                                displayModeBar=True,
                                                modeBarButtonsToRemove=_buttons_to_use('resetScale2d'),
                                            )
                        )),
                        _build_options_btn('clps-dev-bar-button'),
                        dbc.Collapse(id='clps-dev-bar',
                                     style={'paddingLeft': '2rem',
                                            'paddingRight': '2rem',
                                            'paddingBottom': '2rem',
                                            },
                                    children=_option_grid(
                                         labels=['Plot:', 'Scale:', 'Order: '],
                                         values=[
                                            _build_row_radio_button(
                                                rd_id='devs_bar-scale',
                                                options=['linear', 'log'],
                                                value='linear'),
                                            _build_row_radio_button(
                                                rd_id='devs_bar-order',
                                                options=['alphabetical', 'count', 'area'],
                                                value='count'),
                        ]))
            ]),
            dbc.Col(width=6, children=[
                        dbc.Row(dcc.Graph(id='devs_graph-boxplot',
                                          style=dict(height='100%',width='100%'),
                                          figure=fig_bp_state,
                                          config=dict(displaylogo=False,
                                                displayModeBar=True,
                                                modeBarButtonsToRemove=_buttons_to_use(
                                                    'zoom2d', 'pan2d', 'zoomIn2d', 'zoomOut2d',
                                                    'lasso2d', 'resetScale2d'
                                                ),
                                          )
                        )),
                        _build_options_btn('clps-dev-bp-button'),
                        dbc.Collapse(id='clps-dev-bp',
                                     style={'paddingLeft': '2rem',
                                            'paddingRight': '2rem',
                                            'paddingBottom': '2rem',
                                            },
                                    children=_option_grid(
                                         labels=['Plot:', 'Binary state:', ],
                                         lwidth=4, rwidth=8,
                                         values=[
                                            _build_row_radio_button(
                                                rd_id='devs_bp-scale',
                                                options=['linear', 'log'],
                                                value='log'),
                                            _build_row_radio_button(
                                                rd_id='devs_bp-binary-state',
                                                options=['off', 'on'],
                                                value='on'),
                        ]))
            ]),
        ]),
        dbc.Row([
            dbc.Col(width=7,
                children=[
                    dbc.Tabs(id='devs_tab',
                        children=[

                    dbc.Tab(label='D', tab_id='devs_tab-density',
                            label_style={'fontSize': '12px', 'padding': '5px'},
                            children=[
                                dcc.Graph(id='devs_graph-density',
                                          figure=fig_ev_density,
                                          style=dict(height='100%',width='100%'),
                                          config=dict(displaylogo=False,
                                                displayModeBar=True,
                                                modeBarButtonsToRemove=[]
                                          )
                                ),
                                _build_options_btn('clps-dev-density-button'),
                                dbc.Collapse(id='clps-dev-density',
                                         style={'paddingLeft': '2rem',
                                                'paddingRight': '2rem',
                                                'paddingBottom': '2rem',
                                                },
                                         children=_option_grid(
                                             labels=['Resolution', 'Scale'],
                                             lwidth=3, rwidth=9,
                                             values=[
                                                dcc.Slider(id='devs_dens-slider', min=0, max=len(DEV_DENS_SLIDER), step=None, value=5,
                                                           marks={i:{'label':DEV_DENS_SLIDER[i], 'style': {"transform": "rotate(45deg)"}} for i in range(0, len(DEV_DENS_SLIDER))}
                                                ),
                                                 _build_row_radio_button(
                                                      rd_id='devs_dens-scale',
                                                      options=['linear', 'log'],
                                                      value='linear'),
                                            ]
                                        )
                            )

                        ]),
                    ])
                ]
            ),
            dbc.Col(width=5,
                    children=[
                        dbc.Row(style={'marginTop': '29px'},
                            children=dcc.Graph(id='devs_graph-fraction',
                                          style=dict(height='100%',width='100%'),
                                          figure=fig_fraction,
                                          config=dict(displaylogo=False,
                                                displayModeBar=True,
                                                modeBarButtonsToRemove=_buttons_to_use(
                                                    'zoom2d', 'pan2d', 'resetScale2d'
                                              ),
                                            )
                        )),

            ])

        ]),
        dbc.Row(
            children=[
                dbc.Col(width=7, children=[
                            dbc.Row(dcc.Graph(id='devs_graph-cc',
                                            figure=fig_cc,
                                            style=dict(height='100%',width='100%'),
                                            config=dict(displaylogo=False,
                                                    displayModeBar=True,
                                                    modeBarButtonsToRemove=_buttons_to_use('resetScale2d'),
                                                )
                            )),
                            _build_options_btn('clps-dev-cc-button'),
                            dbc.Collapse(id='clps-dev-cc',
                                        style={'paddingLeft': '2rem',
                                                'paddingRight': '2rem',
                                                'paddingBottom': '2rem',
                                                },
                                        children=_option_grid(
                                             labels=['Fix:', 'To:', 'Max lag:', 'Binsize:'],
                                             lwidth=3, rwidth=9,
                                             values=[
                                                _build_choice_devices(df_devs, id='devs_cc-sel-fix', sel=cc_dev_selected),
                                                _build_choice_devices(df_devs, id='devs_cc-sel-to', sel=cc_dev_selected),
                                                dcc.Slider(id='devs_cc-lag-slider', min=0, max=len(LAG_SLIDER), step=None, value=3,
                                                        marks={i:{'label':LAG_SLIDER[i], 'style': {"transform": "rotate(45deg)"}} for i in range(0, len(LAG_SLIDER))}
                                                ),
                                                dcc.Slider(id='devs_cc-binsize-slider', min=0, max=len(BIN_SIZE_SLIDER), step=None, value=3,
                                                        marks={i:{'label':BIN_SIZE_SLIDER[i], 'style': {"transform": "rotate(45deg)"}} for i in range(0, len(BIN_SIZE_SLIDER))}
                                                ),
                                            ]
                                        )
                            )
                ]),
                dbc.Col(width=5, children=[
                            dbc.Row(dcc.Graph(id='devs_graph-iei',
                                          figure=fig_iei,
                                          style=dict(height='100%',width='100%'),
                                          config=dict(displaylogo=False,
                                                displayModeBar=True,
                                                modeBarButtonsToRemove=_buttons_to_use(
                                                        'zoom2d', 'pan2d',  'lasso2d', 'resetScale2d'
                                                ),
                                          )
                                ),
                            ),
                            _build_options_btn('clps-dev-iei-button'),
                            dbc.Collapse(id='clps-dev-iei',
                                        style={'paddingLeft': '2rem',
                                            'paddingRight': '2rem',
                                            'paddingBottom': '2rem',
                                            },
                                    children=_option_grid(
                                            labels=['Scale:', 'Per device:'],
                                            lwidth=4, rwidth=8,
                                            values=[
                                            _build_row_radio_button(
                                                    rd_id='devs_iei-scale',
                                                    options=['linear', 'log'],
                                                    value='log'),
                                                daq.BooleanSwitch(
                                                    id='devs_iei-per-device',
                                                    on=False,
                                                    persisted_props=[]
                            )]))
                ]),
                
            ])

    ]),

    return layout_devices

def _option_grid(labels, values,lwidth=2, rwidth=10):
    """ generates sth. like this:
    lbl_1       *a *b *c
    lbl_2       | asdfasdf v|

    """
    assert (lwidth + rwidth) == 12

    root_child_lst = []
    for lbl, val in zip(labels, values):
        root_child_lst.append(dbc.Row(children=[
            dbc.Col(lbl, width=lwidth),
            dbc.Col(width=rwidth, children=val)
        ]))
    return dbc.Col(children=root_child_lst)

def activities_layout(df_acts, act_id=0, plot_height=350):

    if df_acts is not None and not df_acts.empty:
        fig_bar_count = act_bar_count(df_acts, height=plot_height)
        fig_density = act_density(df_acts, height=plot_height)
        try:
            fig_transition = act_heatmap_transitions(df_acts, height=plot_height)
        except:
            fig_transition = Figure()
        fig_duration = act_boxplot_duration(df_acts, height=plot_height, scale='log')
    else:
        fig_bar_count = Figure()
        fig_density = Figure()
        fig_transition = Figure()
        fig_duration = Figure()

    layout_activities = dbc.Container([
        dcc.Store(f'acts_density-data-{act_id}'),
        dbc.Row([
            dbc.Col(width=6,
                    children=[
                        dbc.Row(dcc.Graph(id=f'acts_graph-bar-{act_id}',
                                          style=dict(height='100%',width='100%'),
                                          figure=fig_bar_count,
                                          config=dict(displaylogo=False,
                                                displayModeBar=True,
                                                modeBarButtonsToRemove=_buttons_to_use(
                                                    'zoom2d', 'pan2d','resetScale2d'
                                              ),
                                            )
                        )),
                        _build_options_btn(f'clps-act-bar-button-{act_id}'),
                        dbc.Collapse(id=f'clps-act-bar-{act_id}',
                                     style={'paddingLeft': '2rem',
                                            'paddingRight': '2rem',
                                            'paddingBottom': '2rem',
                                            },
                                    children=_option_grid(
                                         labels=['Plot:', 'Scale:', 'Order: '],
                                         values=[
                                            dcc.Dropdown(id=f'acts_bar-drop-{act_id}', value='count', options=[
                                                  {'label': 'Count', 'value': 'count'},
                                                  {'label': 'Cumulative', 'value': 'cum'}]
                                            ),
                                            _build_row_radio_button(
                                                rd_id=f'acts_bar-scale-{act_id}',
                                                options=['linear', 'log'],
                                                value='linear'),
                                            _build_row_radio_button(
                                                rd_id=f'acts_sort-{act_id}',
                                                options=['alphabetical', 'value', 'area'],
                                                value='value'),
                        ]))
            ]),
            html.Div(children='', id=f"act-curr-sel-{act_id}", style=dict(display="none")),
            dcc.Store(id=f'act-curr-sel-store-{act_id}'),
            html.Div(id=f"act-trigger-{act_id}", style=dict(display="none"), **{
              "data-value-2": "false"
            }),
            dcc.Store(id=f'acts_activity-order-{act_id}'),
            dbc.Col(width=6,
                    children=[
                        dbc.Row(dcc.Graph(id=f'acts_graph-boxplot-{act_id}',
                                  style=dict(height='100%',width='100%'),
                                  figure=fig_duration,
                                  config=dict(displayModeBar=True,
                                              modeBarButtonsToRemove=_buttons_to_use(
                                                'zoom2d', 'pan2d', 'zoomIn2d', 'zoomOut2d',
                                                'lasso2d', 'resetScale2d'
                                              ),
                                              responsive=False,
                                              displaylogo=False)
                        )),
                        _build_options_btn(f'clps-act-boxplot-button-{act_id}'),
                        dbc.Collapse(id=f'clps-act-boxplot-{act_id}',
                                    style={'paddingLeft': '2rem',
                                            'paddingRight': '2rem',
                                            'paddingBottom': '2rem',
                                            },
                                     children=_option_grid(
                                         labels=['Plot:', 'Scale:'],
                                         values=[
                                            dcc.Dropdown(id=f'acts_bp-drop-{act_id}', value='bp', options=[
                                                                              {'label': 'boxplot', 'value': 'bp'},
                                                                              {'label': 'voilin', 'value': 'vp'}]
                                            ),
                                            _build_row_radio_button(
                                                rd_id=f'acts_bp-scale-{act_id}',
                                                options=['linear', 'log'],
                                                value='log'),
                                         ])
                        ),
            ]),
        ]),
        dbc.Row([
            dbc.Col(width=6,
                    children=[
                        dbc.Row(dcc.Graph(id=f'acts_graph-density-{act_id}',
                                  style=dict(height='100%',width='100%'),
                                  figure=fig_density,
                                  config=dict(
                                        modeBarButtonsToRemove=_buttons_to_use(
                                            'resetScale2d'
                                        ),
                                      displaylogo=False,
                                      displayModeBar=True,
                                      showAxisDragHandles=False,
                                  )),
                        ),
            ]),
            dbc.Col(width=6,
                    children=[
                        dbc.Row(dcc.Graph(id=f'acts_graph-transition-{act_id}',
                                  style=dict(height='100%',width='100%'),
                                  figure=fig_transition,
                                  config=dict(displayModeBar=True,
                                              displaylogo=False,
                                              modeBarButtonsToRemove=_buttons_to_use(
                                                'resetScale2d'
                                                ),
                                  )),
                        ),
                        _build_options_btn(f'clps-act-transition-button-{act_id}'),
                        dbc.Collapse(id=f'clps-act-transition-{act_id}',
                                    style={'paddingLeft': '2rem',
                                            'paddingRight': '2rem',
                                            'paddingBottom': '2rem',
                                            },
                                     children=_option_grid(
                                         labels=['Scale:'],
                                         values=[
                                            _build_row_radio_button(
                                                rd_id=f'acts_trans-scale-{act_id}',
                                                options=['linear', 'log'],
                                                value='linear'),
                                         ]),
                        )
            ]),
        ]),
        dcc.Store(id=f'activity-order-{act_id}'),
    ])
    return layout_activities




import plotly.figure_factory as ff
import plotly.express as px
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from plotly.colors import n_colors
from datetime import timedelta

import numpy as np

from pyadlml.constants import DEVICE, TIME, VALUE, STRFTIME_HOUR, STRFTIME_DATE, PRIMARY_COLOR, SECONDARY_COLOR
from pyadlml.dataset.plot.plotly.activities import _set_compact_title, _scale_xaxis
from pyadlml.dataset.stats.devices import event_count, event_cross_correlogram, event_cross_correlogram3, events_one_day, \
                                          inter_event_intervals
from pyadlml.dataset.util import check_scale, activity_order_by, device_order_by
from pyadlml.dataset.stats.devices import state_times
from plotly.graph_objects import Figure

@check_scale
def bar_count(df_dev, scale='linear', height=350, order='count') -> Figure:
    """ Plots the activities durations against each other
    """
    title ='Event count'
    col_label = 'event_count'
    xlabel = 'log count' if scale == 'log' else 'count'

    df = event_count(df_dev.copy())
    df.reset_index(level=0, inplace=True)

    dev_order = device_order_by(df_dev, rule=order)


    fig = px.bar(df, y=DEVICE,
                 category_orders={DEVICE: dev_order},
                 color_discrete_sequence=[PRIMARY_COLOR],
                 x=col_label,
                 orientation='h')

    _set_compact_title(fig, title)

    fig.update_layout(margin=dict(l=0, r=0, b=0, t=30, pad=0), height=height)
    fig.update_yaxes(title=None)
    fig.update_xaxes(title=xlabel)
    if scale == 'log':
        fig.update_xaxes(type='log')

    return fig

@check_scale
def device_iei(df_devs, scale='linear', height=350, n_bins=20, per_device=False, order='alphabetical') -> Figure:
    """
        plots
    """
    title ='Inter-event-interval'
    col_label = 'event_count'
    xlabel = 'log seconds' if scale == 'log' else 'seconds'
    color = 'device' if per_device else None

    # Get array of seconds
    df = df_devs.copy().sort_values(by=[TIME])
    df['ds'] = df[TIME].diff().shift(-1) / pd.Timedelta(seconds=1)
    if scale == 'log':
        df['ds'] = df['ds'].apply(np.log)

    cds = [PRIMARY_COLOR] if not per_device else None

    fig = px.histogram(df, x='ds', color=color, nbins=n_bins,
                       color_discrete_sequence=cds)

    # TODO refactor, get better second labeling
    hist, bin_edges = np.histogram(df['ds'].values[:-1], bins=n_bins)
    # when n_bins is event than bin_edges is n_bins + 1, therefore
    # TODO critical, make expresseion that fits for general nbins
    mask = (np.tile([0, 1], n_bins//2 + 1)[:-int(n_bins % 2 == 0)] == 1)

    right_bins = bin_edges[mask]
    left_bins = bin_edges[~mask]
    size = right_bins[0] - left_bins[0]
    fig.update_traces(xbins=dict(start=left_bins, size=size))
    #fig.update_traces(xbins_end=right_bins, xbins_start=left_bins, xbins_size=n_bins)

    _set_compact_title(fig, title)
    fig.update_layout(margin=dict(l=0, r=0, b=0, t=30, pad=0), height=height)
    fig.update_yaxes(title=None)
    fig.update_xaxes(title=xlabel)

    # Update the hovertemplate to include the name and custom data
    # and device name
    if per_device:
        for i in range(len(fig.data)):
            fig.data[i].customdata = [fig.data[i].legendgroup]*len(fig.data[i].x)
        fig.update_traces(hovertemplate='device=%{customdata}<br>ds=%{x}<br>count=%{y}<extra></extra>')
    return fig


def fraction(df_dev, height=350, order='alphabetical') -> Figure:
    """
        plots the fraction a device is on vs off over the whole time
    """
    title = 'State fraction'
    xlabel = 'fraction'
    from pyadlml.dataset.stats.devices import state_fractions

    dev_order = device_order_by(df_dev, rule=order)

    # TODO, include td as custom data

    # Returns 'device', 'value', 'td', 'frac'
    df = state_fractions(df_dev)
    def fm1(x):
        if x in ['on', 'off']:
            return x == 'on'
        else:
            return x
    def f(x):
        if isinstance(x, bool):
            return 'on' if x else 'off'
        else:
            return x
    df[VALUE] = df[VALUE].apply(f)

    fig = px.bar(df, y=DEVICE, x='frac', orientation='h', color=VALUE,
                 color_discrete_sequence=[PRIMARY_COLOR, SECONDARY_COLOR],
                 category_orders={DEVICE: dev_order}
                 )

    # Set hovertemplate and custom data
    for i in range(len(fig.data)):
        val = fig.data[i].legendgroup
        mask = df[DEVICE].isin(fig.data[i].y) & (df[VALUE] == val)
        cd = np.array([df.loc[mask, 'td'].astype(str),
                       [fm1(val)]*len(fig.data[i].x)
        ])
        fig.data[i].customdata = np.moveaxis(cd, 0, 1)
        hover_template = 'Device=%{y}<br>Value=' + val + '<br>Fraction=%{x}<br>' + \
                         'Total=%{customdata[0]}<extra></extra>'
        fig.data[i].hovertemplate = hover_template

    _set_compact_title(fig, title)
    fig.update_layout(barmode='stack', margin=dict(l=0, r=0, b=0, t=30, pad=0),
                      height=height)
    fig.update_yaxes(title=None, visible=False, showticklabels=False)
    fig.update_xaxes(title=xlabel)

    return fig


def event_density(df_dev, dt='1h', height=350, scale='linear', show_colorbar=True, order='alphabetical'):
    """
    Computes the heatmap for one day where all the device triggers are showed
    """
    title = 'Event density'
    df = events_one_day(df_dev.copy(), dt)
    time = df['time'].copy()
    df = df.drop(columns='time')
    dev_order = device_order_by(df_dev, rule=order)
    df = df[dev_order]
    vals = df.T.values

    n = len(df)
    devs = df.columns
    dates = pd.Timestamp('01.01.2000 00:00:00') + np.arange(1, n+1) * pd.Timedelta(dt) \
            - pd.Timedelta(dt)/2    # move tick from centered box half dt to left

    if scale == 'log':
        vals = np.log(vals)

    fig = px.imshow(vals, color_continuous_scale='Viridis',
                    y=devs,
                    x=dates,
                    )
    # Create Hoverdata time intervals
    cd = np.tile(dates, (len(df.columns), 1))
    cd = np.array([cd-pd.Timedelta(dt)/2, cd+pd.Timedelta(dt)/2])
    fig['data'][0]['customdata'] = np.moveaxis(cd, 0, -1)


    fig.update_layout(title=title, xaxis_nticks=24, xaxis_tickangle=-45,)
    fig.update_xaxes(tickformat="%H:%M")
    fig.update_layout(margin=dict(l=0, r=0, b=0, t=30, pad=0), height=height)
    _set_compact_title(fig, title)

    if not show_colorbar:
        fig.update_layout(coloraxis_showscale=False)
        fig.update(layout_coloraxis_showscale=False)
        fig.update_coloraxes(showscale=False)

    fig.data[0].hovertemplate = 'Time: %{customdata[0]|' + STRFTIME_HOUR + '}' \
                                + '-%{customdata[1]|' + STRFTIME_HOUR + '}<br>' \
                                + 'Device: %{y}<br>Count: %{z}<extra></extra>'

    return fig



def boxplot_state(df_devs, scale='linear', height=350, binary_state='on',
                  order='alphabetical') -> Figure:
    """ Plot a boxplot of activity durations (mean) max min
    """
    title = 'State distribution'
    xlabel = 'log seconds' if scale == 'log' else 'seconds'

    df = state_times(df_devs, binary_state=binary_state, categorical=True)
    df['seconds'] = df['td']/pd.Timedelta('1s')

    dev_order = device_order_by(df_devs, rule=order)

    # Add column for hover display of datapoints later
    #df[START_TIME] = df_act[START_TIME].dt.strftime('%c')
    #df[END_TIME] = df_act[END_TIME].dt.strftime('%c')
    df['td'] = df['td'].apply(str)

    # TODO refactor, find way to render boxplot points with webgl
    points = 'all' if len(df) < 30000 else 'outliers'

    fig = px.box(df, y=DEVICE, x='seconds', orientation='h',
                 labels=dict(seconds=xlabel),
                 category_orders={DEVICE: dev_order},
                 color_discrete_sequence=[PRIMARY_COLOR],
                 notched=False, points=points,
                 hover_data=['td', 'time']
    )

    _set_compact_title(fig, title)
    if scale == 'log':
        r_min = np.floor(np.log10(df['seconds'].min()))
        r_max = np.ceil(np.log10(df['seconds'].max()))
        fig.update_xaxes(type="log", range=[r_min, r_max])  # log range: 10^r_min=1, 10^r_max=100000


    fig.update_yaxes(title=None, visible=False, showticklabels=False)
    fig.update_layout(margin=dict(l=0, r=0, b=0, t=30, pad=0), height=height)
    fig.data[0].hovertemplate = 'Device=%{y}<br>Seconds=%{x}<br>Td=%{customdata[0]}<br>' +\
                                'Start time=%{customdata[1]|' + STRFTIME_DATE + '}<extra></extra>'

    return fig



def plotly_device_event_correlogram(df_devs, corr_data=(None, None, None), height=600):

    from pyadlml.dataset.stats.devices import event_cross_correlogram2
    if corr_data:
        cc, bins, devs = corr_data
    else:
        cc, bins, devs = event_cross_correlogram2(df_devs, v1=False, use_dask=True)

    fig = make_subplots(rows=len(devs), cols=len(devs), horizontal_spacing=0.005,
                        vertical_spacing=0.005, column_titles=devs.tolist(), row_titles=devs.tolist(),
    
    )
    from pyadlml.dataset.plot.plotly.util import CatColMap

    def plot_histogram(fig, row, col, bins, counts, dev1, dev2, cm):
        counts = np.concatenate([counts[:int(len(counts)/2)],
                            counts[int(len(counts)/2)+1:]]) 
        fig.add_trace(
            go.Bar(
                x=bins,
                y=counts,
                opacity=1,
                hovertemplate=f'Fix {dev1} w.r.t events of {dev2}<extra></extra>',
                showlegend=False,
                marker=dict(line_width=0, color=cm[dev1])
            ), row=row, col=col
        )
        fig.update_xaxes(showticklabels=False, row=row, col=col)
        fig.update_yaxes(showticklabels=False, row=row, col=col)
        
        return fig
    cm = CatColMap()
    for i in range(len(devs)):
        for j in range(len(devs)):
            cm.update(cat=devs[i])
            fig = plot_histogram(fig, i+1,j+1, bins, cc[i, j], devs[i], devs[j], cm)

    # Rotate columns and row annotation correctly 
    for i in range(len(devs)*2):
        if i < len(devs):
            # Case for cols
            fig.layout['annotations'][i]['textangle'] = -90
        else:
            fig.layout['annotations'][i]['textangle'] = 0

    return fig


def plotly_event_correlogram(df_devs=None, cc_data=None, fix=[], to=[], max_lag='2min', binsize='1s', use_dask=False, height=600):

        
        # Wrap if not list
        fix = [fix] if isinstance(fix, str) else fix
        to = [to] if isinstance(to, str) else to
            
        if cc_data is None:
            cc, bins, rows, cols = event_cross_correlogram3(
                df_devs, fix=fix, to=to, maxlag=max_lag, binsize=binsize,
                use_dask=use_dask
            )
        else:
            cc, bins, rows, cols = cc_data[0], cc_data[1], cc_data[2], cc_data[3]

        fig = make_subplots(rows=len(rows), cols=len(cols), horizontal_spacing=0.005,
                            vertical_spacing=0.005, column_titles=cols, row_titles=rows,
        
        )
        from pyadlml.dataset.plot.plotly.util import CatColMap
        import pandas as pd
        def plot_histogram(fig, row, col, bins, counts, dev1, dev2, cm):
            counts = np.concatenate([counts[:int(len(counts)/2)],
                                counts[int(len(counts)/2)+1:]]) 

            tmp = pd.interval_range(start=pd.Timedelta('0s'), end=pd.Timedelta(max_lag), freq=pd.Timedelta(binsize)).values
            len_cd = len(bins)
            cd = np.full(bins.shape, fill_value='', dtype=object)
            for i in range(0, len(cd)):
                if i < len(tmp)-1:
                    cd[i] = '(-' + str(tmp[len(tmp)-i-1].right) + ', -' + str(tmp[len(tmp)-i-1].left) + ']'
                elif i == len(tmp)-1:
                    cd[i] = '(-' + str(tmp[len(tmp)-i-1].right) + ', ' + str(tmp[len(tmp)-i-1].left) + ']'
                elif i == len(tmp):
                    cd[i] = '(0 days 00:00:00, 0 days 00:00:00]'
                else:
                    cd[i] = str(tmp[(i-1)%len(tmp)])
            half = int(np.ceil(len(bins)/2))
            before = np.full(half, fill_value='before', dtype=object)
            after = np.full(half, fill_value='after', dtype=object)
            cd1 = np.concatenate([before, after])[:-1]
            cd1[half-1] = 'at'
            cd = np.stack([cd, cd1]).T 


            fig.add_trace(
                go.Bar(
                    x=bins,
                    y=counts,
                    opacity=1,
                    customdata=cd, 
                    hovertemplate='%{y} events of "' + dev1 + '" occur during %{customdata[0]}<br>%{customdata[1]}' + f' a type "{dev2}" event<extra></extra>',
                    showlegend=False,
                    marker=dict(line_width=0, color=cm[dev2])
                ), row=row, col=col
            )
            fig.update_xaxes(showticklabels=False, row=row, col=col)
            fig.update_yaxes(showticklabels=False, row=row, col=col)

            return fig

        # Color row-wise
        cm = CatColMap()
        for i in range(len(rows)):
            for j in range(len(cols)):
                cm.update(cat=cols[j])
                fig = plot_histogram(fig, i+1,j+1, bins, cc[i, j], rows[i], cols[j], cm)

        # Rotate columns and row annotation correctly 
        for i in range(len(cols) + len(rows)):
            if i < len(cols):
                # Case for cols
                hm = {1:-0, 2:-10, 3:-20, 4:-40, 5:-50}
                #fig.layout['annotations'][i]['textangle'] = hm[len(cols)]
                fig.layout['annotations'][i]['textangle'] = -20
            else:
                hm = {1: 70, 2: 80, 3:80, 4:80, 5:80}
                #fig.layout['annotations'][i]['textangle'] = hm[len(rows)] 
                fig.layout['annotations'][i]['textangle'] = 70

        fig.update_layout(height=height)
        return fig




import pandas as pd
from pyadlml.constants import TIME, DEVICE, VALUE
import datetime

def load_homeassistant(db_url:str, start_time=None, end_time=None) -> pd.DataFrame:
    """
    Loads the Home Assistant database into a pandas dataframe.

    Parameters
    ----------
    db_url : str
        A valid Home Assistant database url. Is used to establish the connection..
    start_time : str, optional
        Datetime string to exclude all rows before the specified start time.
    end_time : str, optional
        Datetime string to exclude all rows after the specified end time.

    Examples
    --------
    >>> from pyadlml.dataset import load_homeassistant
    >>> db_url = "sqlite:///config/homeassistant-v2.db"
    >>> load_homeassistant(db_url)
                    entity_id         state                 last_changed
    0        binary_sensor.p1            on   2019-07-23 07:21:59.954940
    1        binary_sensor.p1           off   2019-07-23 07:21:47.404236
    ...                   ...            ...                         ...
    1199320           sun.sun  above_horizon  2019-05-05 10:35:41.198544
    [1199321 rows x 3 columns]

    Returns
    -------
    df : pd.DataFrame
    """
    limit=5000000

    """ weird new requirement 
    """
    if 'sqlite' in db_url:
        lc_null_fix = ",iif(last_changed is NULL,last_updated,last_changed) as last_changed"
    elif 'mariadb' in db_url:
        lc_null_fix = ",if(last_changed is NULL,last_updated,last_changed) as last_changed"
    elif 'postgresql' in db_url:
        lc_null_fix = ',(case when last_changed is NULL then last_updated else last_changed end)'
    else:
        print("WARNING: LAST_CHANGED may be null not substituted")
        lc_null_fix = ''


    if start_time == None and end_time == None:
        query = f"""
        SELECT entity_id, state, last_updated
        {lc_null_fix}
        FROM states
        WHERE
            state NOT IN ('unknown', 'unavailable')
        ORDER BY last_updated DESC
        LIMIT {limit}
        """
    else:
        if end_time == None:
            end_time = datetime.datetime.utcnow()
        query = f"""
        SELECT entity_id, state, last_updated
        {lc_null_fix}
        FROM states
        WHERE
            state NOT IN ('unknown', 'unavailable') AND
            last_updated BETWEEN 
            '{start_time}' AND '{end_time}'
        ORDER BY last_updated ASC
        LIMIT {limit}
        """
    df = pd.read_sql_query(query, db_url)
    df = df.drop(columns='last_updated')
    return df

def load_homeassistant_devices(db_url, device_list, start_time=None, end_time=None):
    """
    Creates a device dataframe for selected devices within a certain timeframe.

    Parameters
    ----------
    db_url : str
        A valid Home Assistant database url. Is used to establish the connection.
    device_list : lst
        device selection to filter
    start_time : str, optional
        Datetime string to exclude all rows before the specified start time.
    end_time : str, optional
        Datetime string to exclude all rows after the specified end time.

    Examples
    --------
    >>> from pyadlml.dataset import load_homeassistant_devices
    >>> db_url = "sqlite:///config/homeassistant-v2.db"
    >>> lst_dev = ['binary_sensor.b1','switch.computer','light.l1','light.l2']
    >>> df_devices = load_homeassistant(db_url, device_list=lst_dev)

    Returns
    -------
    df_devs : pd.DataFrame
        All recorded devices from the Home Assistant dataset. For more information refer to
        :ref:`user guide<device_dataframe>`.
    """
    
    df = load_homeassistant(db_url, start_time=start_time, end_time=end_time)


    # Bring into pyadlml structure
    df = df[df['entity_id'].isin(device_list)]
    df = df.rename(columns={'entity_id': DEVICE, 'state': VALUE, 'last_changed': TIME})
    df = df[[TIME, DEVICE, VALUE]]

    # Convert to proper datatypes
    df[TIME] = pd.to_datetime(df[TIME])

    # Map binary device values to -> True, False
    # TODO critical, what happens if a non-binary-device 
    #      is on or off or if binary-device is not on or off
    df[VALUE] = df[VALUE].replace(to_replace='on', value=1)
    df[VALUE] = df[VALUE].replace(to_replace='off', value=0)

    return df


import unittest
from hassbrain_algorithm.controller import Controller, Dataset
from hassbrain_algorithm.datasets.pendigit.pendigits import DatasetPendigits
from hassbrain_algorithm.models._model import ModelPendigits

class TestPendigitsModel(unittest.TestCase):
    def setUp(self):
        self.ctrl = Controller()
        self.ctrl.load_dataset(Dataset.PENDIGITS)
        pd_model = ModelPendigits(self.ctrl, "test")
        self.ctrl.register_model(pd_model)
        self.dset = self.ctrl._dataset # type: DatasetPendigits
        self.pd_model = self.ctrl._model

    def test_init_hmms(self):
        self.ctrl.init_model_on_dataset()

    def test_save_hmms(self):
        self.ctrl.init_model_on_dataset()
        self.ctrl.save_model()

    def test_load_hmms(self):
        self.ctrl.load_model()
        # attention manually refresh reference
        self.pd_model = self.ctrl._model

    def test_train_hmms(self):
        self.ctrl.init_model_on_dataset()
        self.ctrl.train_model()

    def test_train_n_save_hmms(self):
        self.ctrl.init_model_on_dataset()
        self.ctrl.train_model()
        self.ctrl.save_model()

    def test_pre_bench_hmms(self):
        self.ctrl.load_model()
        self.pd_model = self.ctrl._model
        y_true, y_pred = self.pd_model.create_pred_act_seqs(self.dset)

    def test_bench_hmms(self):
        self.ctrl.load_model()
        self.pd_model = self.ctrl._model
        self.ctrl.register_benchmark()
        rep = self.ctrl.create_report(
            conf_matrix=True,
            accuracy=True,
            precision=True,
            recall=True,
            f1=True
        )
        print(rep)

    def test_generate_obs(self):
        self.ctrl.load_model()
        self.pd_model = self.ctrl._model
        self.ctrl.generate_observations()

    def test_generate_obs_n_plot(self):
        # numbers that were good in benchmark (desc):
        # [0,8,4,9,3,7]
        self.ctrl.load_model()
        ds = self.ctrl._dataset # type: DatasetPendigits
        pdmod = self.ctrl._model # type: ModelPendigits

        num = 1
        zero_start = [8,1,1]
        one_start = [8,2,1,1,1]
        start_seq =  one_start


        hmm = pdmod._model_dict[num]
        hmm.set_format_full(True)
        print(hmm)

        pdmod.select_number(num)
        seq = pdmod.generate_observations(start_seq)

        ds.plot_obs_seq(seq, num)

    def tearDown(self):
        pass


class TestDatasetPendigits(unittest.TestCase):
    def setUp(self):
        # set of observations
        self.cm = Controller()
        self.cm.load_dataset(Dataset.PENDIGITS)
        self.pd = self.cm._dataset # type: DatasetPendigits

    def tearDown(self):
        pass

    def test_plotting(self):
        self.pd.load_data()
        self.pd.plot_example(12)

    def test_create_train_sequences(self):
        self.pd.load_data()
        enc_data, lengths = self.pd._create_train_seq(1)

    def test_get_double_strokes(self):
        self.pd.load_data()
        digit_data = self.pd._get_train_data_where_double_strokes(9)
        cnt = 0
        for example in digit_data:
            cnt +=1
            self.pd._plotUniPenData(example)
            if cnt == 50:
                break

    def test_train_seq_number(self):
        self.pd.load_data()
        for numn in range(0,40):
            digit_data = self.pd._get_train_data_by_number(numn)
            for ex in range(0, 10):
                print('~'*100)
                example = digit_data[ex+5]
                print(example)
                self.pd._plotUniPenData(example)


    def test_create_test_sequences(self):
        self.pd.load_data()
        enc_data, lengths = self.pd._create_test_seq(0)
        self.assertEqual(59, len(enc_data))
        self.assertEqual(59, lengths[0])

    def test_generate_points_plot_0(self):
        seq = [8,1,1,0,0,0,7,7,6,6,6,5,5,4,4,4,3,3,3,2,2,2,1,1,9,8,1,1,1,1,1,1]
        self.pd.plot_obs_seq(seq, 0)

    def test_generate_points_plot_1(self):
        seq = [8,2,1,1,1,1,1,0,6,5,5,5,5,6,5,9,8,0,0,0,0,0,0,0]
        self.pd.plot_obs_seq(seq, 1)

    def test_generate_points_plot_4(self):
        seq = [8,6,6,6,6,6,6,6,7,7,0,0,1,0,7,0,0,0,9,8,6,6,6,6,6,6,6]
        self.pd.plot_obs_seq(seq, 4)

    def test_generate_points_plot_5(self):
        seq = [8,6,6,6,6,6,7,0,0,0,0,7,6,6,6,6,5,5,5,4,4,4,4,9,8,0,0,0,0,0,0]
        self.pd.plot_obs_seq(seq, 5)

    def test_generate_points_plot_7(self):
        seq = [8,0,0,0,0,7,6,5,5,5,5,5,5,5,5,9,8,0,0,0,0,0,0]
        self.pd.plot_obs_seq(seq, 7)

    def test_generate_points_plot_8(self):
        seq = [8,2,1,1,0,0,7,7,6,6,6,5,5,4,4,3,3,2,9,8,2,1,1,0,0,7,7,6,6,5,5,4,4,3,3,2]
        self.pd.plot_obs_seq(seq, 8)

    def test_generate_points_plot_9(self):
        seq = [8,3,4,5,6,6,7,0,0,1,1,2,2,3,3,4,4,5,6,9,8,7,6,6,6,6,5,5,5,4,4,3]
        self.pd.plot_obs_seq(seq, 9)

    def test_directions_to_points(self):
        stepsize = 5
        self.pd._resolution = 8
        x0 = self.pd._new_point_x(0, 0, stepsize)
        y0 = self.pd._new_point_y(0, 0, stepsize)
        x1 = self.pd._new_point_x(0, 1, stepsize)
        y1 = self.pd._new_point_y(0, 1, stepsize)

        x2 = self.pd._new_point_x(0, 2, stepsize)
        y2 = self.pd._new_point_y(0, 2, stepsize)
        x3 = self.pd._new_point_x(0, 3, stepsize)
        y3 = self.pd._new_point_y(0, 3, stepsize)

        x4 = self.pd._new_point_x(0, 4, stepsize)
        y4 = self.pd._new_point_y(0, 4, stepsize)
        x5 = self.pd._new_point_x(0, 5, stepsize)
        y5 = self.pd._new_point_y(0, 5, stepsize)

        x6 = self.pd._new_point_x(0, 6, stepsize)
        y6 = self.pd._new_point_y(0, 6, stepsize)
        x7 = self.pd._new_point_x(0, 7, stepsize)
        y7 = self.pd._new_point_y(0, 7, stepsize)

        self.assertEqual(5.000, round(x0, 3))
        self.assertEqual(0.000, round(y0, 3))
        self.assertEqual(3.536, round(x1, 3))
        self.assertEqual(3.536, round(y1, 3))

        self.assertEqual(0.000, round(x2, 3))
        self.assertEqual(5.000, round(y2, 3))
        self.assertEqual(-3.536, round(x3, 3))
        self.assertEqual(3.536, round(y3, 3))

        self.assertEqual(-5.000, round(x4, 3))
        self.assertEqual(0.000, round(y4, 3))
        self.assertEqual(-3.536, round(x5, 3))
        self.assertEqual(-3.536, round(y5, 3))

        self.assertEqual(0.000, round(x6, 3))
        self.assertEqual(-5.000, round(y6, 3))
        self.assertEqual(3.536, round(x7, 3))
        self.assertEqual(-3.536, round(y7, 3))


    def test_points_to_direction(self):
        # directions 0 - 7
        # number of classes the direction can have
        c = 8

        ## ---- 0 degree
        direc = self.pd._points_to_direction(c, 0, 0, 1, 0)
        self.assertEqual(0, direc)
        ## ---- 45 degree
        direc = self.pd._points_to_direction(c, 0, 0, 1, 1)
        self.assertEqual(1, direc)
        ## ---- 90 degree
        direc = self.pd._points_to_direction(c, 0, 0, 0, 1)
        self.assertEqual(2, direc)
        ## ---- 135 degree
        direc = self.pd._points_to_direction(c, 0, 0, -1, 1)
        self.assertEqual(3, direc)
        ## ---- 180 degree
        direc = self.pd._points_to_direction(c, 0, 0, -1, 0)
        self.assertEqual(4, direc)
        ## ---- 225 degree
        direc = self.pd._points_to_direction(c, 0, 0, -1, -1)
        self.assertEqual(5, direc)
        ## ---- 270 degree
        direc = self.pd._points_to_direction(c, 0, 0, 0, -1)
        self.assertEqual(6, direc)
        ## ---- 315 degree
        direc = self.pd._points_to_direction(c, 0, 0, 1, -1)
        self.assertEqual(7, direc)

        # random other angles
        # ----  52 degree
        direc = self.pd._points_to_direction(c, 0, 0, 0.61, 0.79)
        self.assertEqual(1, direc)
        ## ---- 100 degree
        direc = self.pd._points_to_direction(c, 0, 0, -0.18, 0.98)
        self.assertEqual(2, direc)
        # ---- 291 degree
        direc = self.pd._points_to_direction(c, 0, 0, 0.36, -0.93)
        self.assertEqual(6, direc)
        # ----- 350 degree
        direc = self.pd._points_to_direction(c, 0, 0, 0.98, -0.17)
        self.assertEqual(0, direc)

    def test_points_to_direction_2(self):
        c = 8
        # check if direction correct after pen is set on table or
        # after pen is removed from table
        xp = 0
        yp = 0
        # ----- 350 degree
        direc = self.pd._points_to_direction(c, xp, yp, 0.98, -0.17)
        self.assertEqual(0, direc)



from pathlib import Path
from pyadlml.constants import OTHER, START_TIME, ACTIVITY, END_TIME
import pandas as pd
import numpy as np

"""
    df_activities:
        - per definition no activity can be performed in parallel

        start_time | end_time   | activity
        ---------------------------------
        timestamp   | timestamp | act_name


"""
INT_EPS = pd.Timedelta('1ms')
COLS = [START_TIME, END_TIME, ACTIVITY]

def read_activity_csv(path_to_file: Path) -> pd.DataFrame:
    activities = pd.read_csv(path_to_file)
    activities[START_TIME] = pd.to_datetime(activities[START_TIME])
    activities[END_TIME] = pd.to_datetime(activities[END_TIME])
    return activities


def write_activity_csv(path_to_file: Path, df: pd.DataFrame) -> None:
    # TODO string format the timestamps
    df.to_csv(path_to_file, index=False)



def get_index_matching_rows(df_act, rows, tolerance='1ms'):
    """

    Note
    ----
    Attention, since rows are read from strings or whatever the format assumed is dayfirst
               i.e. 01.08.2009 equalst the first of august.
    
    """
    if isinstance(rows, list):
        df = pd.DataFrame(rows, columns=COLS)
        df[START_TIME] = pd.to_datetime(df[START_TIME], errors='coerce', dayfirst=True)
        df[END_TIME] = pd.to_datetime(df[END_TIME], errors='coerce', dayfirst=True)
    else:
        print('went here')
        df = rows

    assert isinstance(df, pd.DataFrame)

    tol = pd.Timedelta(tolerance)
    idxs = []
    for idx, row in df.iterrows():
        mask_st = (row[START_TIME]-tol < df_act[START_TIME])\
                & (df_act[START_TIME] < row[START_TIME]+tol)
        mask_et = (row[END_TIME]-tol < df_act[END_TIME])\
                & (df_act[END_TIME] < row[END_TIME]+tol)
        mask_act = (df_act[ACTIVITY] == row[ACTIVITY])
        res = df_act[mask_st & mask_et & mask_act].index.values
        assert len(res) <= 1
        if len(res) == 1:
            idxs.append(*res)
        if len(res) == 0:
            print('Warning!!!. Activity corrections are not applied!')
    return idxs



def is_activity_df(df):
    """
        :param: df
        start_time | end_time   | activity
        ----------------------------------
        timestamp   | timestamp | act_name
    """
    return START_TIME in df.columns \
        and END_TIME in df.columns \
        and ACTIVITY in df.columns \
        and len(df.columns) == 3


def check_activity_df(df):
    """
    check if the activitiy dataframe is valid by checking if
        - the dataframe has the correct dimensions and labels
        - activities are non overlapping

    :param: df 
    start_time | end_time   | activity
    ----------------------------------
    timestamp   | timestamp | act_name

    """
    if not is_activity_df(df):
        raise ValueError
    if _is_activity_overlapping(df):
        print('there should be none activity overlapping')
        raise ValueError
    return True


def _is_activity_overlapping(df):
    """ checks if any activity is overlapping another
    Parameters
    ----------
    df : pd.DataFrame

    """
    assert df.shape[1] == 3, "Activity dataframes must have 3 columns."

    df = df.sort_values(by=START_TIME).reset_index(drop=True)
    mask = (df[END_TIME].shift()-df[START_TIME]) >= pd.Timedelta('0ms')
    overlapping = df[mask]
    return not overlapping.empty


def create_empty_activity_df():
    """
    returns: empty pd Dataframe 
    """
    df = pd.DataFrame(columns=[START_TIME, END_TIME, ACTIVITY])
    df[START_TIME] = pd.to_datetime(df[START_TIME])
    df[END_TIME] = pd.to_datetime(df[END_TIME])
    return df


def add_other_activity(acts, min_diff=pd.Timedelta('5s')):
    """ adds a dummy Idle activity for gaps between activities greater than min_diff
    Parameters
    ----------
    acts: pd.DataFrame
        the activity data with columns: start_time, end_time, activity
    Returns
    ----------
    pd.DataFrame
        activity data with columns: start_time, end_time, activity
    """
    acts = acts.copy().reset_index(drop=True)

    def func(series, df):
        """
        """
        # check if at end of the series
        if series.name == len(df)-1:
            return series
        else:
            next_entry = df.loc[series.name+1, :]
            return pd.Series({
                START_TIME: series.end_time + pd.Timedelta('1ms'),
                END_TIME: next_entry.start_time - pd.Timedelta('1ms'),
                ACTIVITY: OTHER
            })

    acts['diff'] = acts[START_TIME].shift(-1) - acts[END_TIME]
    tmp = acts[acts['diff'] > min_diff].copy()
    tmp = tmp.apply(func, axis=1, df=acts)
    res = pd.concat([acts.iloc[:, :-1], tmp]).sort_values(by=START_TIME)
    return res.reset_index(drop=True)


def _merge_int_inclusive(row, ov, strat='inplace'):
    """
    int |~~~~~~~~|
    ov    |----|   => |~|----|~~|

    Parameters
    ----------
    ov: pd.DataFrame
        single df with one row
    row: pd.Series
        one row 
    """
    assert strat in ['inplace']
    df = create_empty_activity_df()

    if strat == 'inplace':
        df.loc[0] = [row.start_time, ov.start_time, row.activity]
        df.loc[1] = [ov.start_time + INT_EPS, ov.end_time, ov.activity]
        df.loc[2] = [ov.end_time + INT_EPS, row.end_time, row.activity]
    else:
        raise ValueError('Strategy not known.')
    return df


def _merge_int_right_partial(row, ov, strat='clip_left'):
    """ merges 
        row int |~~~~|    => |~~~|---|    
        ov ints   |----|  
    """
    df_res = create_empty_activity_df()

    # get row stats
    row_act = row[ACTIVITY]
    row_st = row[START_TIME]
    row_et = row[END_TIME]

    # get overlap stats
    ov_st = ov.start_time
    ov_et = ov.end_time
    ov_act = ov.activity

    if strat == 'clip_left':
        df_res.loc[0] = [row_st, ov_st, row_act]
        df_res.loc[1] = [ov_st + INT_EPS, ov_et, ov_act]
    else:
        raise ValueError('Strategy not known.')

    return df_res


def _merge_int_first_persists(row1, row2):
    """ Replaces interval by taking the first as dominant interval which borders
        should be preserved:
            1. case 
                row1 int |~~~~|    => |~~~~|-|    
                row2 int   |----|  
            2. case
                row1 int   |~~~~|  => |-|~~~~|    
                row2 int |----|

    Parameters
    ----------
    row1 : pd.Series
        The dominant interval
    row2 : pd.Series
        the submissive interval

    Returns
    -------
    df_res : pd.DataFrame
        the corrected activities
    """
    df = create_empty_activity_df()
    eps = pd.Timedelta('1ms')

    int1 = pd.Interval(row1.start_time, row1.end_time)
    int2 = pd.Interval(row2.start_time, row2.end_time)

    if int1.left <= int2.left and int1.right < int2.right:  # 1.case
        df.loc[0] = [row1.start_time, row1.end_time, row1.activity]
        df.loc[1] = [row1.end_time + eps, row2.end_time, row2.activity]

    elif int2.left <= int1.left and int2.right < int1.right:    # 2.case
        df.loc[0] = [row2.start_time, row1.start_time, row2.activity]
        df.loc[1] = [row1.start_time + eps, row1.end_time, row1.activity]
    return df


def _merge_int_same(row1, row2):
    """ Two intervals of the same type are merged into one.
       row1 int    |~~~~|  => |~~~~~~|
       row2 int |~~~~|

    Parameters
    ----------
    row1 : pd.Series
    row2 : pd.Series

    Returns
    -------
    df_res : pd.DataFrame
        the corrected activities
    """
    assert row1.activity == row2.activity

    df = create_empty_activity_df()
    df.loc[0] = [row1.start_time, row2.end_time, row1.activity]
    return df


def _merge_ints(row, overlapping, strats=['cut_at_lower'], excepts=[]):
    """ gets overlapping intervals and merges those intervals with a strategy
    Parameters
    ----------
    row : pd.Series 

    overlapping : pd.Series 

    strats : list
        the strategies for merging intervals with priority in ascending order
    excepts : list
        the activities which should be preserved when merging

    Returns
    -------
    merged: pd.DataFrame
    """
    # todo check if every strategy in the list is possible
    assert isinstance(overlapping, pd.Series)
    assert isinstance(row, pd.Series)

    if row.activity == overlapping.activity:
        return _merge_int_same(row, overlapping)

    int1 = pd.Interval(row.start_time, row.end_time)
    int2 = pd.Interval(overlapping.start_time, overlapping.end_time)

    if excepts != []:
        # find out which is the dominant and the less dominant interval
        if row.activity in excepts and overlapping.activity in excepts:
            # assign row and ov such as to replace the one with lower priority below
            print('priority mismatch!'*20)
            idx_row = excepts.index(row.activity)
            idx_ov = excepts.index(overlapping.activity)
            if idx_row < idx_ov:
                dom = overlapping
                less_dom = row

        if row.activity in excepts:
            dom = row
            less_dom = overlapping
        else:
            dom = overlapping
            less_dom = row

        # apply merging strategies
        int_dom = pd.Interval(dom.start_time, dom.end_time)
        int_ldom = pd.Interval(less_dom.start_time, less_dom.end_time)

        # dominant interval encloses less dominant => keep dominant, drop less dominant
        if (int_dom.left < int_ldom.left) & (int_ldom.right < int_dom.right):
            df_res = create_empty_activity_df()
            df_res.loc[0] = dom
            return df_res

        # less dominant interval encloses dominant => normal inclusive merge
        elif (int_ldom.left < int_dom.left) & (int_dom.right < int_ldom.right):
            return _merge_int_inclusive(less_dom, dom)

        # intervals overlap => keep dominant
        else:
            return _merge_int_first_persists(dom, less_dom)

    if (int1.left < int2.left) & (int2.right < int1.right):
        # int1  |~~~~~~|
        # int2   |----|
        df_res = _merge_int_inclusive(row, overlapping)

    elif (int1.left <= int2.left) & (int1.right < int2.right):
        # int1 |~~~~|
        # int2   |----|
        df_res = _merge_int_right_partial(row, overlapping)

    else:
        raise ValueError  # this should never happen
    return df_res


def _is_activity_et_after_st(df):
    """ checks if the start_time is lesser than end_time for every activity
    """
    df = df.sort_values(by=START_TIME)
    df = df.copy()

    df['diff'] = df['end_time'] - df['start_time']
    mask = df['diff'] < pd.Timedelta('0ns')
    return not mask.any()


def correct_activity_overlap(df_act, strategies=[], excep=[]):
    """ solve the merge overlapping interval problem
        worst runtime is O(n^2)
        average runtime is O()

    Parameters
    ----------
    df_act : pd.DataFrame
        Activity dataframe with the columns
    strategies : list, default=[]
        the strategies for merging intervals with priority in ascending order
    excep : list, default=[]
        the activities which should be preserved when merging

    Returns
    -------
    df : pd.DataFrame 
        corrected activity dataframe
    corrections : list
        a list of tuples with the areas that had to be corrected and the corrections
    """
    df_act = df_act.copy()\
                   .sort_values(START_TIME)\
                   .reset_index(drop=True)

    res = create_empty_activity_df()
    corrections = []

    # get all activities that have an overlap with a direct preceding interval
    mask = (df_act[START_TIME].shift(-1) -
            df_act[END_TIME]) <= pd.Timedelta('0ms')
    idxs_succ_overlaps = np.array(list(df_act[mask].index))

    i_l, i = 0, 0
    while i < len(idxs_succ_overlaps):
        # Append correct parts up to another overlap to result
        i_h = idxs_succ_overlaps[i]
        # exclusive i_h e,g 0->2 is [0,1]
        res = pd.concat([res, df_act.iloc[i_l:i_h, :]])

        # Get index of first element where start_time is lesser than the end_time
        # this marks the point where we can copy again indices
        i_l = i_h
        i_h = _first_non_overlapped_int_after_idx(df_act, i_h)

        # only for last iteration
        if i == len(idxs_succ_overlaps)-1:
            area_to_correct = df_act.iloc[i_l:i_h, :]
            result = _correct_overlapping_segment(
                area_to_correct, strategies, excep)
            corrections.append((area_to_correct, result))
            res = pd.concat([res, result])
            break

        # If the last overlapped interval exceeds any pair of overlapping indices, extent the range to
        # include them and adjust the loop iteration
        if (i_h >= idxs_succ_overlaps[i+1:]).any():
            """ 
            for all succ ov indicies get the maximum of the extented range
            e.g max-int = 4
               1 |~~~~~~~~~~~~~~|
               2          |--|
               3              |~~~~| 
               4                 |---|
               5                       |---| <----- max-int = 5
               6                            |~~~~|
               7                              |-|
            """
            for idx in idxs_succ_overlaps[i+1:]:
                if idx >= i_h:
                    break
                # If the succeeding pair is in the range of the overlap skip the next outer loop iteration
                # and set the upper overlap limit to the maximum of the included rows
                i += 1
                i_h = max(_first_non_overlapped_int_after_idx(df_act, idx), i_h)
        area_to_correct = df_act.iloc[i_l:i_h, :].copy()
        result = _correct_overlapping_segment(
            area_to_correct, strategies, excep)

        assert not result.empty
        corrections.append((area_to_correct, result))
        res = pd.concat([res, result])
        i += 1
        i_l = i_h

    res = pd.concat([res, df_act.iloc[i_h:, :]])

    # sanity checks
    assert len(res) >= len(df_act)
    added_entrys = 0
    for corr in corrections:
        added_entrys += len(corr[1]) - len(corr[0])
    assert len(res) - added_entrys == len(df_act)

    res = res.sort_values(by=START_TIME)
    res = res.reset_index(drop=True)
    return res, corrections


def _first_non_overlapped_int_after_idx(df, idx):
    """ Returns the index of the first element in the dataframe where the
    start_time is greater than the end_time of the interval that overlaps 
    the succeeding one. Thus marking the overlapping areas that need correction.
        idx=3 |----------------------|
        4         |~~~~~~|
        5                  |++++++++++++++++|
        6                               |~~~~|
        => Returns 6

    Parameters
    ----------
    df : pd.DataFrame
        activity dataframe
    idx : int
        index of the row where the interval overlaps the succeeding row(s)

    Returns
    -------
    res : int
        index of a row
    """
    return list(df[df[START_TIME] > df.iloc[idx, :].end_time].index)[0]


def _correct_overlapping_segment(area_to_correct, strats, excepts=[]):
    """
    Parameters
    ----------
    area_to_correct : pd.DataFrame
        Acitity dataframe
    strats : list
        the strategies for merging intervals with priority in ascending order
    excepts : list
        the activities which should be preserved when merging

    """
    assert len(area_to_correct) >= 2, f'Test {area_to_correct}'

    result = create_empty_activity_df()
    stack = area_to_correct.copy()\
                           .sort_values(by=START_TIME)\
                           .reset_index(drop=True)

    while True:
        # pop first and second item from stack if they overlap otherwise
        # append to result until two items overlap
        while True:
            current_row = stack.iloc[0]
            ov = stack.iloc[1]

            # if they don't overlap push onto result otherwise proceed with merging
            # procedure
            int1 = pd.Interval(current_row.start_time, current_row.end_time)
            int2 = pd.Interval(ov.start_time, ov.end_time)
            if int1.overlaps(int2) or current_row.end_time == ov.start_time:
                stack = stack.iloc[2:]  # remove first element from stack
                break
            # the case when the current two activities of interest are the last ones
            elif stack.iloc[2:].empty:
                result = pd.concat([result, stack])
                return result
            else:
                result = pd.concat([result, current_row.to_frame().T], ignore_index=True)
                stack = stack.iloc[1:]

        new_rows = _merge_ints(current_row, ov, strats, excepts)
        
        assert isinstance(new_rows, pd.DataFrame)

        if stack.empty:
            result = pd.concat([result, new_rows], axis=0, ignore_index=True)
            return result
        elif len(new_rows) >= 2:
            result = pd.concat([result, new_rows.iloc[0,:].to_frame().T], ignore_index=True)
            new_rows = new_rows.iloc[1:]

        stack = pd.concat([stack, new_rows], axis=0, ignore_index=True)
        stack = stack.sort_values(by=START_TIME)
    return result


def _get_overlapping_activities(df, shift=1):
    """ gets all activities that have an overlap
    """
    assert shift >= 1

    df = df.copy()
    df = df.sort_values(by=START_TIME)
    df = df.reset_index(drop=True)

    # get all activities that are have in an overlap
    mask = (df[START_TIME].shift(-shift) - df[END_TIME]) < pd.Timedelta('0ms')

    # as start_time is shifted upwards to select the right corresp. overlap
    # shift the mask 'shift' steps downards
    mask = mask.shift(+shift) | mask
    return df[mask]


def exists_st_before_et(df_acts):
    """

    """
    df = df_acts.copy()
    df['diff'] = df[END_TIME] - df[START_TIME]
    mask = (df['diff'] < pd.Timedelta('0s'))
    violating = df[mask]
    return not violating.empty

def correct_succ_same_end_and_start_time(df: pd.DataFrame) -> pd.DataFrame:
    """ Correct pairwise activities where the firsts end_time is equal to the seconds start_time
        by adding a ms onto the respective start_time

    Parameters
    ----------
    df : pd.DataFrame
        Activity dataframe

    """
    test = df[END_TIME].shift(1) - df[START_TIME]
    mask = (test == pd.Timedelta('0s'))
    df.loc[mask, START_TIME] += pd.Timedelta(INT_EPS)
    return df

def correct_activities(df, strats=[], excepts=[], retain_corrections=False):
    """ gets df in form of activities and removes overlapping activities
    Parameters
    ----------
    df : pd.DataFrame
    strats : list
        the strategies for merging intervals with priority in ascending order
    excepts : list
        the activities which should be preserved when merging
    retain_corrections: boolean, default=False
        Whether the applied corrections should be returned for examination

    Returns
    -------
    df : pd.DataFrame
    corrections: list
    """
    corrections = []
    df = df.copy()
    df = df.drop_duplicates(ignore_index=True)

    # Correct pairwise activities where the firsts end_time is equal to the seconds start_time
    df = correct_succ_same_end_and_start_time(df)

    # Check if an end_time is greater than a start_time
    if exists_st_before_et(df):
        raise NotImplementedError(
            'ST > ET violation. This has to be hard fixed, since there is no simple heuristic.')

    # Correct overlapping activities
    if _is_activity_overlapping(df):
        df, corrections = correct_activity_overlap(df, strats, excepts)
    assert not _is_activity_overlapping(df)

    if retain_corrections:
        return df, corrections
    else:
        return df, None


class ActivityDict(dict):
    """ Dictionary with activity pd.DataFrames as values and subject names as keys.
    """

    def subjects(self) -> list:
        return list(self.keys())

    def to_json(self, date_unit="ns"):
        """Serialize to json"""
        tmp = {}
        for k, df in self.items():
            tmp[k] = df.to_json(date_unit=date_unit)

        return json.dumps(tmp)

    def read_json(cls, string):        
        """Serialize from json"""
        tmp = json.loads(string)
        for k, str in tmp.items():
            tmp[k] = pd.read_json(str)
        return ActivityDict(tmp)

    def nr_acts(self):
        """"""
        return max([len(df_acts[ACTIVITY].unique()) for df_acts in self.values()])

    def get_activity_union(self): 
        return list(set([item for v in self.values() \
                              for item in v[ACTIVITY].unique()]))

    def apply(self, func):
        """ Applies a function to each dataframe
        """
        for k, df in self.items():
            self[k] = func(df)
        return self

    def min_starttime(self):
        min_lst = []
        for df_acts in self.values():
            if not df_acts.empty:
                min_lst.append(df_acts[START_TIME].iloc[0]) 
        return min(min_lst)

    def max_endtime(self):
        max_lst = []
        for df_acts in self.values():
            if not df_acts.empty:
                max_lst.append(df_acts[END_TIME].iloc[-1]) 
        return max(max_lst)

    def concat(self):
        return pd.concat(self.values())

    @classmethod
    def wrap(cls, df_acts):
        if isinstance(df_acts, pd.DataFrame): 
            df_acts = df_acts.copy().reset_index(drop=True)  # TODO not here
            df_acts = ActivityDict({'subject':df_acts})
            return df_acts
        elif isinstance(df_acts, list):
            return ActivityDict({f'subject_{i}':df for i, df in enumerate(df_acts)})
        elif isinstance(df_acts, ActivityDict):
            return df_acts
        elif isinstance(df_acts, dict):
            return ActivityDict(df_acts)
        else:
            raise NotImplementedError

    def unwrap(self, inst_type: type):
        if inst_type  == ActivityDict:
            return self
        elif inst_type == list:
            return list(self.values())
        elif inst_type == dict:
            return super(self)
        elif inst_type == pd.DataFrame:
            assert len(self) == 1
            return list(self.values())[0]
        else:
            raise NotImplementedError





''' Define the Layers '''
import torch.nn as nn
import torch
from transformer.SubLayers import MultiHeadAttention, PositionwiseFeedForward


__author__ = "Yu-Hsiang Huang"


class EncoderLayer(nn.Module):
    ''' Compose with two layers '''

    def __init__(self, d_model, d_inner, n_head, d_k, d_v, dropout=0.1):
        super(EncoderLayer, self).__init__()
        self.slf_attn = MultiHeadAttention(n_head, d_model, d_k, d_v, dropout=dropout)
        self.pos_ffn = PositionwiseFeedForward(d_model, d_inner, dropout=dropout)

    def forward(self, enc_input, slf_attn_mask=None):
        enc_output, enc_slf_attn = self.slf_attn(
            enc_input, enc_input, enc_input, mask=slf_attn_mask)
        enc_output = self.pos_ffn(enc_output)
        return enc_output, enc_slf_attn


class DecoderLayer(nn.Module):
    ''' Compose with three layers '''

    def __init__(self, d_model, d_inner, n_head, d_k, d_v, dropout=0.1):
        super(DecoderLayer, self).__init__()
        self.slf_attn = MultiHeadAttention(n_head, d_model, d_k, d_v, dropout=dropout)
        self.enc_attn = MultiHeadAttention(n_head, d_model, d_k, d_v, dropout=dropout)
        self.pos_ffn = PositionwiseFeedForward(d_model, d_inner, dropout=dropout)

    def forward(
            self, dec_input, enc_output,
            slf_attn_mask=None, dec_enc_attn_mask=None):
        dec_output, dec_slf_attn = self.slf_attn(
            dec_input, dec_input, dec_input, mask=slf_attn_mask)
        dec_output, dec_enc_attn = self.enc_attn(
            dec_output, enc_output, enc_output, mask=dec_enc_attn_mask)
        dec_output = self.pos_ffn(dec_output)
        return dec_output, dec_slf_attn, dec_enc_attn



import unittest
import numpy as np
import os
from matplotlib import pyplot as plt
import copy, os

import pyhsmm
from matplotlib.figure import Figure
from pyhsmm.util.text import progprint_xrange
import ssm

class TestController(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass


    def test_hmm(self):
        T = 100  # number of time bins
        K = 5    # number of discrete states
        D = 2    # dimension of the observations

        # make an hmm and sample from it
        hmm = ssm.HMM(K, D, observations="gaussian")
        z, y = hmm.sample(T)
        print(z)
        print(y)

        #Fitting an HMM is simple.

        test_hmm = ssm.HMM(K, D, observations="gaussian")
        test_hmm.fit(y)
        zhat = test_hmm.most_likely_states(y)
        print(zhat)

    def test_hsmm_example(self):
        import autograd.numpy as np
        import autograd.numpy.random as npr
        from scipy.stats import nbinom
        import matplotlib.pyplot as plt
        import ssm
        from ssm.util import rle, find_permutation

        npr.seed(0)

        # Set the parameters of the HMM
        T = 5000    # number of time bins
        K = 5       # number of discrete states
        D = 2       # number of observed dimensions

        # Make an HMM with the true parameters
        true_hsmm = ssm.HSMM(K, D, observations="gaussian")
        print(true_hsmm.transitions.rs)
        z, y = true_hsmm.sample(T)
        z_test, y_test = true_hsmm.sample(T)
        true_ll = true_hsmm.log_probability(y)

        # Fit an HSMM
        N_em_iters = 500

        print("Fitting Gaussian HSMM with EM")
        hsmm = ssm.HSMM(K, D, observations="gaussian")
        hsmm_em_lls = hsmm.fit(y, method="em", num_em_iters=N_em_iters)

        print("Fitting Gaussian HMM with EM")
        hmm = ssm.HMM(K, D, observations="gaussian")
        hmm_em_lls = hmm.fit(y, method="em", num_em_iters=N_em_iters)

        # Plot log likelihoods (fit model is typically better)
        plt.figure()
        plt.plot(hsmm_em_lls, ls='-', label="HSMM (EM)")
        plt.plot(hmm_em_lls, ls='-', label="HMM (EM)")
        plt.plot(true_ll * np.ones(N_em_iters), ':', label="true")
        plt.legend(loc="lower right")

        # Print the test likelihoods (true model is typically better)
        print("Test log likelihood")
        print("True HSMM: ", true_hsmm.log_likelihood(y_test))
        print("Fit HSMM:  ", hsmm.log_likelihood(y_test))
        print("Fit HMM: ", hmm.log_likelihood(y_test))

        # Plot the true and inferred states
        hsmm.permute(find_permutation(z, hsmm.most_likely_states(y)))
        hsmm_z = hsmm.most_likely_states(y)
        hmm.permute(find_permutation(z, hmm.most_likely_states(y)))
        hmm_z = hsmm.most_likely_states(y)


        # Plot the true and inferred discrete states
        plt.figure(figsize=(8, 6))
        plt.subplot(311)
        plt.imshow(z[None, :1000], aspect="auto", cmap="cubehelix", vmin=0, vmax=K-1)
        plt.xlim(0, 1000)
        plt.ylabel("True $z")
        plt.yticks([])

        plt.subplot(312)
        plt.imshow(hsmm_z[None, :1000], aspect="auto", cmap="cubehelix", vmin=0, vmax=K-1)
        plt.xlim(0, 1000)
        plt.ylabel("HSMM Inferred $z$")
        plt.yticks([])

        plt.subplot(313)
        plt.imshow(hmm_z[None, :1000], aspect="auto", cmap="cubehelix", vmin=0, vmax=K-1)
        plt.xlim(0, 1000)
        plt.ylabel("HMM Inferred $z$")
        plt.yticks([])
        plt.xlabel("time")

        plt.tight_layout()

        # Plot the true and inferred duration distributions
        states, durations = rle(z)
        inf_states, inf_durations = rle(hsmm_z)
        hmm_inf_states, hmm_inf_durations = rle(hmm_z)
        max_duration = max(np.max(durations), np.max(inf_durations), np.max(hmm_inf_durations))
        dd = np.arange(max_duration, step=1)

        plt.figure(figsize=(3 * K, 9))
        for k in range(K):
            # Plot the durations of the true states
            plt.subplot(3, K, k+1)
            plt.hist(durations[states == k] - 1, dd, density=True)
            plt.plot(dd, nbinom.pmf(dd, true_hsmm.transitions.rs[k], 1 - true_hsmm.transitions.ps[k]),
                     '-k', lw=2, label='true')
            if k == K - 1:
                plt.legend(loc="lower right")
            plt.title("State {} (N={})".format(k+1, np.sum(states == k)))

            # Plot the durations of the inferred states
            plt.subplot(3, K, K+k+1)
            plt.hist(inf_durations[inf_states == k] - 1, dd, density=True)
            plt.plot(dd, nbinom.pmf(dd, hsmm.transitions.rs[k], 1 - hsmm.transitions.ps[k]),
                     '-r', lw=2, label='hsmm inf.')
            if k == K - 1:
                plt.legend(loc="lower right")
            plt.title("State {} (N={})".format(k+1, np.sum(inf_states == k)))

                # Plot the durations of the inferred states
            plt.subplot(3, K, 2*K+k+1)
            plt.hist(hmm_inf_durations[hmm_inf_states == k] - 1, dd, density=True)
            plt.plot(dd, nbinom.pmf(dd, 1, 1 - hmm.transitions.transition_matrix[k, k]),
                     '-r', lw=2, label='hmm inf.')
            if k == K - 1:
                plt.legend(loc="lower right")
            plt.title("State {} (N={})".format(k+1, np.sum(hmm_inf_states == k)))
        plt.tight_layout()

        plt.show()


    def test_own_hsmm_example(self):
        import autograd.numpy as np
        import autograd.numpy.random as npr
        from scipy.stats import nbinom
        import matplotlib.pyplot as plt
        import ssm
        from ssm.util import rle, find_permutation

        print(npr.seed(0))

        # Set the parameters of the HMM
        T = 1000     # number of time bins todo why can't I set this < 500
        K = 8       # number of discrete states
        D = 5       # number of observed dimensions

        # Make an HMM with the true parameters
        true_hsmm = ssm.HSMM(K, D, observations="categorical")
        z, y = true_hsmm.sample(T)
        z_test, y_test = true_hsmm.sample(T)
        true_ll = true_hsmm.log_probability(y)

        # Fit an HSMM
        N_em_iters = 100

        print("Fitting Categorical HSMM with EM")
        hsmm = ssm.HSMM(K, D, observations="categorical")
        hsmm_em_lls = hsmm.fit(y, method="em", num_em_iters=N_em_iters)

        print("Fitting Categorical HMM with EM")
        hmm = ssm.HMM(K, D, observations="categorical")
        hmm_em_lls = hmm.fit(y, method="em", num_em_iters=N_em_iters)

        # Plot log likelihoods (fit model is typically better)
        plt.figure()
        plt.plot(hsmm_em_lls, ls='-', label="HSMM (EM)")
        plt.plot(hmm_em_lls, ls='-', label="HMM (EM)")
        plt.plot(true_ll * np.ones(N_em_iters), ':', label="true")
        plt.legend(loc="lower right")

        # Print the test likelihoods (true model is typically better)
        print("Test log likelihood")
        print("True HSMM: ", true_hsmm.log_likelihood(y_test))
        print("Fit HSMM:  ", hsmm.log_likelihood(y_test))
        print("Fit HMM: ", hmm.log_likelihood(y_test))

        # Plot the true and inferred states
        tmp1 = hsmm.most_likely_states(y)
        tmp2 = find_permutation(z, tmp1)
        hsmm.permute(tmp2)
        hsmm_z = hsmm.most_likely_states(y)

        # calculates viterbi sequence of states
        tmp3 = hmm.most_likely_states(y)
        #
        """
        z = true state seq [1,2,1,....,]
        tmp3 = pred. state seq [3,4,1,2,...,]
        match each row to different column in such a way that corresp
        sum is minimized
        select n el of C, so that there is exactly one el.  in each row 
        and one in each col. with min corresp. costs 
        
        
        match states [1,2,...,] of of the 
        """
        tmp4 = find_permutation(z, tmp3)
        hmm.permute(tmp4)
        hmm_z = hsmm.most_likely_states(y)

        # Plot the true and inferred discrete states
        plt.figure(figsize=(8, 6))
        plt.subplot(311)
        plt.imshow(z[None, :1000], aspect="auto", cmap="cubehelix", vmin=0, vmax=K-1)
        plt.xlim(0, 1000)
        plt.ylabel("True $z")
        plt.yticks([])

        plt.subplot(312)
        plt.imshow(hsmm_z[None, :1000], aspect="auto", cmap="cubehelix", vmin=0, vmax=K-1)
        plt.xlim(0, 1000)
        plt.ylabel("HSMM Inferred $z$")
        plt.yticks([])

        plt.subplot(313)
        plt.imshow(hmm_z[None, :1000], aspect="auto", cmap="cubehelix", vmin=0, vmax=K-1)
        plt.xlim(0, 1000)
        plt.ylabel("HMM Inferred $z$")
        plt.yticks([])
        plt.xlabel("time")

        plt.tight_layout()


        # Plot the true and inferred duration distributions
        """
        N = the number of infered states 
            how often the state was inferred 
            blue bar is how often when one was in that state it endured x long
        x = maximal duration in a state
        
        
        red binomial plot
            for the hmm it is 1 trial and the self transitioning probability
            for the hsmm it is
            
        """
        """
        Negativ binomial distribution for state durations
        
            NB(r,p)
                r int, r>0
                p = [0,1] always .5 wk des eintretens von erfolgreicher transition
                r = anzahl erflogreiche selbst transitionen  befor man etwas anderes (trans in anderen
                zustand sieht)
                
                
        
        
        """

        true_states, true_durations = rle(z)
        hmm_inf_states, hmm_inf_durations = rle(hmm_z)
        hsmm_inf_states, hsmm_inf_durations = rle(hsmm_z)
        max_duration = max(np.max(true_durations), np.max(hsmm_inf_durations), np.max(hmm_inf_durations))
        max_duration = 100
        dd = np.arange(max_duration, step=1)

        plt.figure(figsize=(3 * K, 9))
        for k in range(K):
            # Plot the durations of the true states
            plt.subplot(3, K, k+1)
            """
            get the durations where it was gone into the state k =1
            state_seq: [0,1,2,3,1,1]
            dur_seq: [1,4,5,2,4,2]
                meaning one ts in state 0, than 4 in state 1, 5 in state 2, so on and so forth
            x = [4,4,2]
            """
            x = true_durations[true_states == k] - 1
            plt.hist(x, dd, density=True)
            n = true_hsmm.transitions.rs[k]
            p = 1 - true_hsmm.transitions.ps[k]
            plt.plot(dd, nbinom.pmf(dd, n, p),
                     '-k', lw=2, label='true')
            if k == K - 1:
                plt.legend(loc="lower right")
            plt.title("State {} (N={})".format(k+1, np.sum(true_states == k)))

            # Plot the durations of the inferred states of hmm
            plt.subplot(3, K, 2*K+k+1)
            plt.hist(hmm_inf_durations[hmm_inf_states == k] - 1, dd, density=True)
            plt.plot(dd, nbinom.pmf(dd, 1, 1 - hmm.transitions.transition_matrix[k, k]),
                     '-r', lw=2, label='hmm inf.')
            if k == K - 1:
                plt.legend(loc="lower right")
            plt.title("State {} (N={})".format(k+1, np.sum(hmm_inf_states == k)))

            # Plot the durations of the inferred states of hsmm
            plt.subplot(3, K, K+k+1)
            plt.hist(hsmm_inf_durations[hsmm_inf_states == k] - 1, dd, density=True)
            plt.plot(dd, nbinom.pmf(dd, hsmm.transitions.rs[k], 1 - hsmm.transitions.ps[k]),
                     '-r', lw=2, label='hsmm inf.')
            if k == K - 1:
                plt.legend(loc="lower right")
            plt.title("State {} (N={})".format(k+1, np.sum(hsmm_inf_states == k)))

        plt.tight_layout()

        plt.show()




import pandas as pd
from pyadlml.constants import TIME, DEVICE, VALUE

def create_changepoint(df_devs):
    """

    Parameters
    ----------
    df_devs : pd.DataFrame
        A device dataframe, refer to guide

    Returns
    -------
    df : pd.DataFrame
    """

    # create binary vectors and set all changepoints to true
    df = df_devs.copy()
    df[VALUE] = True
    df = df.pivot(index=TIME, columns=DEVICE, values=VALUE)\
        .fillna(False)\
        .astype(int)\
        .reset_index()
    return df


def resample_changepoint(cp, dt):
    """
    Resamples the changepoint representation with a given resolution

    Parameters
    ----------
    cp : pd.DataFrame
        A device dataframe in changepoint representation

    dt : str

    Returns
    -------
    cp : pd.DataFrame
        Resampled dataframe in changepoint representation
    """
    cp = cp.set_index(TIME)
    resampler = cp.resample(dt, kind='timestamp')
    cp = resampler.apply(_cp_evaluator)\
        .reset_index()
    return cp


def _cp_evaluator(series: pd.Series):
    """ for each element return 1 if there was a change in the given interval or 0 otherwise
    Parameters
    ----------
        series: pd.Series
            contains name of the column the evaluator operates on
            contains the timestamps for that change in set frame and the value 
            that the specified column has at set timestamp.

            is empty if all columns that have no changing entity

        df: pd.DataFrame
            contains device representation 3. Is for determining the states of other
            device/columns for the timestamps

    Returns: 0|1
    """

    if series.empty or series.sum() == 0:
        return 0
    else:
        return 1


import sys
import pathlib

from sklearn.neighbors import KernelDensity
from sklearn.svm import SVC

working_directory = pathlib.Path().absolute()
script_directory = pathlib.Path(__file__).parent.absolute()
sys.path.append(str(working_directory))
import unittest
from sklearn.ensemble import RandomForestClassifier
from pyadlml.dataset import set_data_home, load_act_assist
from pyadlml.pipeline import Pipeline
from pyadlml.preprocessing import StateVectorEncoder, LabelMatcher, DropTimeIndex
SUBJECT_ADMIN_NAME = 'admin'


class TestPipeline(unittest.TestCase):
    def setUp(self):
        dataset_dir = str(script_directory) + '/datasets/partial_dataset'
        self.data = load_act_assist(dataset_dir, subjects=[SUBJECT_ADMIN_NAME])
        self.data.df_activities = self.data.df_activities_admin
        self.rf_pipe = [
            ('raw', StateVectorEncoder(encode='raw')),
            ('lbl', LabelMatcher(other=True)),
            ('drop_tidx', DropTimeIndex()),
            ('classifier', RandomForestClassifier(random_state=42))
        ]
        self.kde_pipe = [
            ('raw', StateVectorEncoder(encode='raw')),
            ('drop_tidx', DropTimeIndex()),
            ('classifier', KernelDensity(kernel='gaussian', bandwidth=0.5))
        ]
        self.svm_pipe = [
            ('raw', StateVectorEncoder(encode='raw')),
            ('lbl', LabelMatcher(other=True)),
            ('drop_tidx', DropTimeIndex()),
            ('classifier', SVC())
        ]


    #def test_decision_function(self):
    #    #  Apply transforms, and decision_function of the final estimator
    #    pipe = Pipeline(self.svm_pipe)
    #    pipe = pipe.fit(self.data.df_devices,
    #                self.data.df_activities)
    #    tmp = pipe.decision_function(self.data.df_devices)


    def test_fit(self):
        pipe = Pipeline(self.rf_pipe)
        tmp = pipe.fit(self.data.df_devices,
                       self.data.df_activities)

    def test_fit_predict(self):
        # Applies fit_predict of last step in pipeline after transforms.
        pipe = Pipeline(self.rf_pipe)
        tmp = pipe.fit_predict(
            self.data.df_devices,
            self.data.df_activities)
        print(tmp)

    def test_fit_transform(self):
        #Fit the model and transform with the final estimator
        pipe = Pipeline(self.rf_pipe)
        tmp = pipe.fit_transform(
            self.data.df_devices,
            self.data.df_activities)


    def test_get_params(self):
        #Get parameters for this estimator.
        pipe = Pipeline(self.rf_pipe)
        tmp = pipe.get_params()
        #print(tmp)


    def test_predict(self):
        #Apply transforms to the data, and predict with the final estimator
        pipe = Pipeline(self.rf_pipe).fit(self.data.df_devices, self.data.df_activities)
        tmp = pipe.predict(self.data.df_devices)
        print(tmp)

    def test_predict_log_proba(self):
        #Apply transforms, and predict_log_proba of the final estimator
        pipe = Pipeline(self.rf_pipe).fit(self.data.df_devices, self.data.df_activities)
        tmp = pipe.predict_log_proba(self.data.df_devices)
        print(tmp)


    def test_predict_proba(self):
        #Apply transforms, and predict_proba of the final estimator
        pipe = Pipeline(self.rf_pipe).fit(self.data.df_devices, self.data.df_activities)
        tmp = pipe.predict_proba(self.data.df_devices)
        print(tmp)


    def test_score(self):
        #Apply transforms, and score with the final estimator
        pipe = Pipeline(self.rf_pipe).fit(self.data.df_devices, self.data.df_activities)
        tmp = pipe.score(
            self.data.df_devices,
            self.data.df_activities)
        print(tmp)

    def test_score_samples(self):
        #Apply transforms, and score_samples of the final estimator.
        pipe = Pipeline(self.kde_pipe).fit(self.data.df_devices)
        tmp = pipe.score_samples(self.data.df_devices)
        print(tmp)

if __name__ == '__main__':
    unittest.main()


import pyadlml.dataset._datasets.activity_assistant as act_assist
from pyadlml.constants import ACTIVITY
from pyadlml.dataset.io.downloader import MegaDownloader
from pyadlml.dataset.io.remote import DataFetcher

TUE_2019_URL = 'https://mega.nz/file/sRIT0ILI#us-EtWRCMtvzoqkbsz8UofAbqomn3Px3CLXw1NxSCxY'
TUE_2019_FILENAME = 'tuebingen_2019.zip'
TUE_2019_CLEANED_URL = 'https://mega.nz/file/wcRhWayA#itY_OorjDdU60RCwY4WMCansb3GqPqvzb6R1o3crNs0'
TUE_2019_CLEANED_FILENAME = 'tuebingen_2019_cleaned.zip'


# The activity and device corrections are already applied from act_assist.load
def fetch_tuebingen_2019(keep_original=False, cache=True, load_cleaned=False,\
                         retain_corrections=False, folder_path=None):
    """
    Fetches the tuebingen_2019 dataset from the internet. The original dataset or its cached version
    is stored in the data_home folder.

    Parameters
    ----------
    keep_original : bool, default=True
        Determines whether the original dataset is deleted after downloading
        or kept on the hard drive.
    cache : bool, default=True
        Determines whether the data object should be stored as a binary file for faster
        succeeding access.
        For more information how caching is used refer to the :ref:`user guide <storage>`.
    retain_corrections : bool, optional, default=False
        When set to *true* data points that are changed or dropped during preprocessing
        are listed in the respective attributes of the data object.  Fore more information
        about the attributes refer to the :ref:`user guide <error_correction>`.

    Returns
    -------
    data : object
    """

    class TueFetcher(DataFetcher):
        def load_data(self, folder_path):
            return act_assist.load(folder_path=folder_path)


    downloader = MegaDownloader(
        url=TUE_2019_URL,
        fn=TUE_2019_FILENAME,
        url_cleaned=TUE_2019_CLEANED_URL,
        fn_cleaned=TUE_2019_CLEANED_FILENAME,
    )

    data_fetch = TueFetcher(
        dataset_name='tuebingen_2019',
        downloader=downloader,
        correct_activities=True,
        correct_devices=True
    )

    
    return data_fetch(keep_original=keep_original, cache=cache, load_cleaned=load_cleaned,
            retain_corrections=retain_corrections, folder_path=folder_path)



import hashlib
import shutil
import pandas as pd
from pathlib import Path


def clear_data_home():
    """ Delete all content inside the data home folder.
    """
    data_home = get_data_home()
    _delete_data(data_home)
    _create_folder(data_home)

def _move_files_to_parent_folder(path_to_folder):
    """ Moves all files in given folder on level up and deletes the empty directory
    """
    import shutil
    source_dir = Path(path_to_folder)
    for file_name in source_dir.iterdir():
        shutil.move(str(source_dir.joinpath(file_name)), source_dir.parent)

    source_dir.rmdir()

def set_data_home(path_to_folder):
    """
    Sets the global variable ``data_home`` and creates the according folder.
    All dump, load and fetch operations assume this folder as the base of operation.

    Parameters
    ----------
    path_to_folder : str
        Specifies the path where the data_home is located.
    """
    from pyadlml import DATA_HOME
    global DATA_HOME
    DATA_HOME = Path(path_to_folder)


def _ensure_dh_folder_exists():
    path_to_folder = get_data_home()
    if not path_to_folder.exists():
        _create_folder(path_to_folder)


def get_data_home() -> Path:
    """ Returns the current folder where pyadlml saves all datasets to.

    Returns
    -------
    path_to_data_home : str
    """
    global DATA_HOME
    return DATA_HOME


def load(name):
    """ Loads a python object from the data_home folder if it exists.

    Parameters
    ----------
    param_dict : dict
        A dictionary identifying the object that is to be loaded. The keys and
        values have to exactly match the objects keys and values when it was
        dumped with *dump_in_data_home*.

    Examples
    --------

    Returns
    -------
    X : pd.DataFrame
        Some observations
    y : pd.DataFrame
        Some labels
    """
    # create folder name
    folder_name = hashdict2str({'str': name})
    folder_path = Path(get_data_home()).joinpath(folder_name)

    if not folder_path.exists():
        raise EnvironmentError(f'Nothing is saved in: {folder_path}')

    res = []
    for f in folder_path.glob('*'):
        df = pd.read_parquet(f)
        res.append(df)
   
    return res


def dump(dfs: list, name: str) -> None:
    """ Creates a folder inside the *data_home* and dumps the given dataframe(s) inside.

    Parameters
    ----------
    dfs : list or pd.DataFrame
        The dataframe or dataframes to save
    name : str
        A string used to generate the folder name where the dataframes are stored.

    """
    dfs = dfs if isinstance(dfs, list) else [dfs]

    _ensure_dh_folder_exists()

    # create string representation of dictionary
    folder_name = hashdict2str({'str': name})
    folder_path = Path(get_data_home()).joinpath(folder_name)

    # remove folder if it already exists
    if folder_path.exists():
        shutil.rmtree(folder_path)
        
    _create_folder(folder_path)

    # save all data
    for i, df in enumerate(dfs):
        fp = folder_path.joinpath(f'df_{i}.parquet')
        #df.to_parquet(fp)
        import pyarrow as pa
        import pyarrow.parquet as pq
        table = pa.Table.from_pandas(df)
        pq.write_table(table, fp)


def _create_folder(path_to_folder):
    Path(path_to_folder).mkdir(parents=True, exist_ok=True)


def hashdict2str(param_dict):
    """ creates a unique string for a dictionary 
    Parameters
    ----------
    param_dict : dict
        Contains the attributes of the encoder and the data representations
    Returns
    -------
    folder_name : str
    """
    # sort dictionary after keys for determinism
    param_dict = {k: v for k, v in sorted(param_dict.items(), key=lambda item: item[1])}

    # create string
    param_string = str(param_dict).encode('utf-8')
    folder_name = hashlib.md5(param_string).hexdigest()
    return str(folder_name)


def _delete_data(path_to_folder):
    # make sure only data in home directory are deleted
    #assert '/home/' == path_to_folder[:6]
    shutil.rmtree(path_to_folder)


def _data_2_folder_name(path_to_folder, data_name):
    param_dict = {'dataset': data_name}
    folder_name = hashdict2str(param_dict)
    folder_path = Path(path_to_folder).joinpath(folder_name)
    return folder_path


import math
from hassbrain_algorithm.models._model import Model
import autograd.numpy as np
import ssm
from ssm.util import rle, find_permutation
from hassbrain_algorithm.models.util import create_xset_onebitflip

class BernoulliHMM(Model):

    def __init__(self, controller):
        #self._cm = controller # type: Controller
        # training parameters
        self._training_steps = 50
        self._epsilon = None
        self._use_q_fct = False

        Model.__init__(self, "test", controller)

    def __str__(self):
        if self._hmm is None:
            return "hmm has to be inits"
        else:
            s = "--"*10 + "\n"
            s += "States: " + str(self._hmm.K) + "\n"
            s += "Obs:\t " + str(self._hmm.D) + "\n"
            s += "pi:\t " + str(self._hmm.init_state_distn) + "\n"
            s += "Trans:\t " + str(self._hmm.transitions) + "\n"
            s += "M:\t " + str(self._hmm.M) + "\n"
            return s

    def set_training_steps(self, val):
        assert val > 0 and val < 1000 and isinstance(val, int)
        self._training_steps = val

    def _model_init(self, dataset):
        K = len(self._state_list)       # number of discrete states
        D = len(self._observation_list) # dimension of the observation
        self._hmm = ssm.HMM(K, D, observations='bernoulli')

    def save_visualization(self, path_to_file):
        """
        save a visualization of the model to the given filepath
        :param path_to_file:
        :return:
        """
        raise NotImplementedError

    def predict_state_sequence(self, test_y):
        pred_x = self._hmm.most_likely_states(test_y)
        return pred_x

    def predict_obs_sequence(self, test_y):
        D, _ = test_y.shape
        pred_y = np.zeros((D), dtype=np.int64)
        #for i in range(D):
        #    self.
        #    self._hmm.predict_xnp1()
        #    tmp =''
        #    res_y[i] = tmp
        _, pred_y = self._hmm.sample(test_y)
        return pred_y

    def can_predict_next_obs(self):
        return True

    def _sample(self, n, obs_seq=None):
        tmp1, tmp2 = self._hmm.sample(T=n)
        return tmp1, tmp2

    def can_predict_prob_devices(self):
        return True

    def draw(self, act_retrieval_meth):
        self._hmm.plot()
        return self._hmm.generate_graphviz_dot_ext_lbl(act_retrieval_meth)
        #vg.render('test.gv', view=True)

    def _train_loss_callback(self, hmm, loss, *args):
        # todo in log models convert to normal Probability
        """
        :param hmm:
        :param loss:
        :param args:
        :return:
        """
        # this is due to the the loss param is actually the likelihood of the P(X|Model)
        # therefore the loss can be 1 - P(X|Model)
        loss = 1-loss
        self._bench.train_loss_callback(hmm, loss)

    def _train(self, dataset):
        y_train = dataset.get_train_data()
        hmm_em_lls = self._hmm.fit(
            y_train,
            method="em",
            num_em_iters=self._training_steps)

        test_x, test_y = dataset.get_all_labeled_data()
        self._assign_states(test_x, test_y)
        return hmm_em_lls

    def _assign_states(self, true_z, true_y):
        """
        assigns the unordered hidden states of the trained model (on true_y)
        to the most probable state labels in alignment of true_z
        :param true_z
            the true state sequence of a labeled dataset
        :param true_y
            the true corresp. observation sequence of a labeled dataset

        assign
           z = true state seq [1,2,1,....,]
           tmp3 = pred. state seq [3,4,1,2,...,]
           match each row to different column in such a way that corresp
           sum is minimized
           select n el of C, so that there is exactly one el.  in each row
           and one in each col. with min corresp. costs


        match states [1,2,...,] of of the
        :return:
            None
        """
        # Plot the true and inferred states
        tmp1 = self._hmm.most_likely_states(true_y)
        # todo temporary cast to int64 remove for space saving solution
        # todo as the amount of states only in range [0, 30
        true_z = true_z.astype(np.int64)
        tmp2 = find_permutation(true_z, tmp1)
        self._hmm.permute(tmp2)

#-------------------------------------------------------------------
    # RT Node stuff

    def _classify(self, obs_seq):
        """
        get the most probable state/activity
        :param obs_seq:
        :return:
        """
        assert len(obs_seq[0]) == self._hmm.D
        state_seq = self._hmm.most_likely_states(obs_seq)
        pred_state = state_seq[-1:][0]
        return pred_state

    def _classify_multi(self, obs_seq):
        """
        computes the last omega slice of viterbi which is
        equivalent to
        :param obs_seq:
        :return:
            numpy array with each index being the id for the label
        """
        tmp = self._hmm.filter(obs_seq)
        last_alpha = tmp[-1:][0]


        assert math.isclose(last_alpha.sum(), 1.)

        K = self._hmm.K
        return last_alpha

    def _predict_next_obs(self, obs_seq):
        """
        the index of the most probable next observation
        """
        samples = create_xset_onebitflip(obs_seq[-1:][0])
        log_prob_states = self._hmm.predict_xnp1(data=obs_seq,x=samples)
        # get max of pre_states
        max_idx = np.argmax(log_prob_states)
        if max_idx == len(log_prob_states) -1:
            return np.argmax(log_prob_states[:len(log_prob_states)-1])
        else:
            return max_idx

    def _predict_prob_xnp1(self, obs_seq):
        samples = create_xset_onebitflip(obs_seq[-1:][0])
        log_prob_states = self._hmm.predict_xnp1(data=obs_seq,x=samples)
        normalizer = ssm.util.logsumexp(log_prob_states)
        norm_log_prob = log_prob_states - normalizer
        norm_log_prob = norm_log_prob[:len(norm_log_prob)-1]
        norm_log_prob = np.exp(norm_log_prob)
        return norm_log_prob
